// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: messages.proto
// Protobuf C++ Version: 7.35.0

#include "messages.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::AstroidMessages::Focus
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Ack
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Indent
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::AllowRemoteImages
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Navigate
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Mark
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Hidden
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Debug
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Page
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Info
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Address
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::AddressList
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Message_Chunk_Signature
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Message_Chunk_Encryption
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Message_Chunk
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::Message
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::State_MessageState_Element
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::State_MessageState
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::State
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::UpdateMessage
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
        // ::AstroidMessages::ClearMessage
        {&::_pbi::kDescriptorMethods, &::descriptor_table_messages_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace AstroidMessages {
class State_MessageState_Element::_Internal {
 public:
  using HasBits = decltype(::std::declval<State_MessageState_Element>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_._has_bits_);
};

constexpr State_MessageState_Element::ParseTableT_ State_MessageState_Element::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::State_MessageState_Element>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool focusable = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(State_MessageState_Element, _impl_.focusable_), 3>(),
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.focusable_)}},
      // .AstroidMessages.State.MessageState.Element.Type type = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(State_MessageState_Element, _impl_.type_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.type_)}},
      // int32 id = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(State_MessageState_Element, _impl_.id_), 2>(),
       {16, 2, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.id_)}},
      // string sid = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.sid_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .AstroidMessages.State.MessageState.Element.Type type = 1;
      {PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.type_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
      // int32 id = 2;
      {PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // string sid = 3;
      {PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.sid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool focusable = 4;
      {PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.focusable_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\52\0\0\3\0\0\0\0"
      "AstroidMessages.State.MessageState.Element"
      "sid"
    }},
  };
}


inline constexpr State_MessageState_Element::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        sid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        type_{static_cast< ::AstroidMessages::State_MessageState_Element_Type >(0)},
        id_{0},
        focusable_{false} {}

template <typename>
constexpr State_MessageState_Element::State_MessageState_Element(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL State_MessageState_Element::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) State_MessageState_Element(arena);
}
constexpr auto State_MessageState_Element::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(State_MessageState_Element), alignof(State_MessageState_Element));
}
constexpr auto State_MessageState_Element::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &State_MessageState_Element::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<State_MessageState_Element>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &State_MessageState_Element::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<State_MessageState_Element>(), &State_MessageState_Element::ByteSizeLong,
              &State_MessageState_Element::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[16],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct State_MessageState_ElementGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr State_MessageState_ElementGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 State_MessageState_Element_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(State_MessageState_Element::InternalGenerateClassData_(
            _default, &State_MessageState_Element_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<State_MessageState_Element>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~State_MessageState_ElementGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) State_MessageState_Element _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<State_MessageState_Element>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(State_MessageState_ElementGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST State_MessageState_ElementGlobalsTypeInternal State_MessageState_Element_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* State_MessageState_Element_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return State_MessageState_Element_globals_.GetClassData();
#else
  return State_MessageState_Element_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Page::_Internal {
 public:
  using HasBits = decltype(::std::declval<Page>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Page, _impl_._has_bits_);
};

constexpr Page::ParseTableT_ Page::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Page, _impl_._has_bits_),
      0, // no _extensions_
      8, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967040,  // skipmap
      offsetof(ParseTableT_, field_entries),
      8,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Page>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string log_level = 8;
      {::_pbi::TcParser::FastUS1,
       {66, 4, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.log_level_)}},
      // string html = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.html_)}},
      // string css = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 2, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.css_)}},
      // string part_css = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 3, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.part_css_)}},
      // repeated string allowed_uris = 4;
      {::_pbi::TcParser::FastUR1,
       {34, 0, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.allowed_uris_)}},
      // bool use_stdout = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Page, _impl_.use_stdout_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.use_stdout_)}},
      // bool use_syslog = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Page, _impl_.use_syslog_), 6>(),
       {48, 6, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.use_syslog_)}},
      // bool disable_log = 7;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Page, _impl_.disable_log_), 7>(),
       {56, 7, 0,
        PROTOBUF_FIELD_OFFSET(Page, _impl_.disable_log_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string html = 1;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.html_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string css = 2;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.css_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string part_css = 3;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.part_css_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated string allowed_uris = 4;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.allowed_uris_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
      // bool use_stdout = 5;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.use_stdout_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool use_syslog = 6;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.use_syslog_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool disable_log = 7;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.disable_log_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // string log_level = 8;
      {PROTOBUF_FIELD_OFFSET(Page, _impl_.log_level_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\24\4\3\10\14\0\0\0\11\0\0\0\0\0\0\0"
      "AstroidMessages.Page"
      "html"
      "css"
      "part_css"
      "allowed_uris"
      "log_level"
    }},
  };
}


inline constexpr Page::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        allowed_uris_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Page,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.allowed_uris_)>()
         }
        ,
        html_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        css_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        part_css_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        log_level_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        use_stdout_{false},
        use_syslog_{false},
        disable_log_{false} {}

template <typename>
constexpr Page::Page(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Page::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Page(arena);
}
constexpr auto Page::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Page), alignof(Page));
}
constexpr auto Page::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Page::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Page>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Page::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Page>(), &Page::ByteSizeLong,
              &Page::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Page, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[8],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct PageGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr PageGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Page_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Page::InternalGenerateClassData_(
            _default, &Page_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Page>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~PageGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Page _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Page>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(PageGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST PageGlobalsTypeInternal Page_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Page_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Page_globals_.GetClassData();
#else
  return Page_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Navigate::_Internal {
 public:
  using HasBits = decltype(::std::declval<Navigate>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Navigate, _impl_._has_bits_);
};

constexpr Navigate::ParseTableT_ Navigate::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Navigate, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Navigate>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // .AstroidMessages.Navigate.Direction direction = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Navigate, _impl_.direction_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(Navigate, _impl_.direction_)}},
      // .AstroidMessages.Navigate.Type type = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Navigate, _impl_.type_), 2>(),
       {16, 2, 0,
        PROTOBUF_FIELD_OFFSET(Navigate, _impl_.type_)}},
      // string mid = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(Navigate, _impl_.mid_)}},
      // int32 element = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Navigate, _impl_.element_), 3>(),
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(Navigate, _impl_.element_)}},
      // bool focus_top = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Navigate, _impl_.focus_top_), 4>(),
       {40, 4, 0,
        PROTOBUF_FIELD_OFFSET(Navigate, _impl_.focus_top_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // .AstroidMessages.Navigate.Direction direction = 1;
      {PROTOBUF_FIELD_OFFSET(Navigate, _impl_.direction_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
      // .AstroidMessages.Navigate.Type type = 2;
      {PROTOBUF_FIELD_OFFSET(Navigate, _impl_.type_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
      // string mid = 3;
      {PROTOBUF_FIELD_OFFSET(Navigate, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // int32 element = 4;
      {PROTOBUF_FIELD_OFFSET(Navigate, _impl_.element_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // bool focus_top = 5;
      {PROTOBUF_FIELD_OFFSET(Navigate, _impl_.focus_top_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\30\0\0\3\0\0\0\0"
      "AstroidMessages.Navigate"
      "mid"
    }},
  };
}


inline constexpr Navigate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        direction_{static_cast< ::AstroidMessages::Navigate_Direction >(0)},
        type_{static_cast< ::AstroidMessages::Navigate_Type >(0)},
        element_{0},
        focus_top_{false} {}

template <typename>
constexpr Navigate::Navigate(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Navigate::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Navigate(arena);
}
constexpr auto Navigate::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Navigate), alignof(Navigate));
}
constexpr auto Navigate::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Navigate::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Navigate>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Navigate::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Navigate>(), &Navigate::ByteSizeLong,
              &Navigate::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Navigate, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[4],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct NavigateGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr NavigateGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Navigate_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Navigate::InternalGenerateClassData_(
            _default, &Navigate_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Navigate>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~NavigateGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Navigate _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Navigate>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(NavigateGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST NavigateGlobalsTypeInternal Navigate_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Navigate_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Navigate_globals_.GetClassData();
#else
  return Navigate_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Message_Chunk_Signature::_Internal {
 public:
  using HasBits = decltype(::std::declval<Message_Chunk_Signature>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_._has_bits_);
};

constexpr Message_Chunk_Signature::ParseTableT_ Message_Chunk_Signature::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk_Signature>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // bool verified = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk_Signature, _impl_.verified_), 2>(),
       {8, 2, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.verified_)}},
      // repeated string sign_strings = 2;
      {::_pbi::TcParser::FastUR1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.sign_strings_)}},
      // repeated string all_errors = 3;
      {::_pbi::TcParser::FastUR1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.all_errors_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bool verified = 1;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.verified_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // repeated string sign_strings = 2;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.sign_strings_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
      // repeated string all_errors = 3;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_.all_errors_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
    }},
    // no aux_entries
    {{
      "\47\0\14\12\0\0\0\0"
      "AstroidMessages.Message.Chunk.Signature"
      "sign_strings"
      "all_errors"
    }},
  };
}


inline constexpr Message_Chunk_Signature::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        sign_strings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Signature,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.sign_strings_)>()
         }
        ,
        all_errors_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Signature,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.all_errors_)>()
         }
        ,
        verified_{false} {}

template <typename>
constexpr Message_Chunk_Signature::Message_Chunk_Signature(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Message_Chunk_Signature::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Message_Chunk_Signature(arena);
}
constexpr auto Message_Chunk_Signature::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Message_Chunk_Signature), alignof(Message_Chunk_Signature));
}
constexpr auto Message_Chunk_Signature::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Message_Chunk_Signature::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Message_Chunk_Signature>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Message_Chunk_Signature::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Message_Chunk_Signature>(), &Message_Chunk_Signature::ByteSizeLong,
              &Message_Chunk_Signature::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Message_Chunk_Signature, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[12],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct Message_Chunk_SignatureGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr Message_Chunk_SignatureGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Message_Chunk_Signature_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Message_Chunk_Signature::InternalGenerateClassData_(
            _default, &Message_Chunk_Signature_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk_Signature>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~Message_Chunk_SignatureGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Message_Chunk_Signature _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk_Signature>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(Message_Chunk_SignatureGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST Message_Chunk_SignatureGlobalsTypeInternal Message_Chunk_Signature_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Message_Chunk_Signature_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Message_Chunk_Signature_globals_.GetClassData();
#else
  return Message_Chunk_Signature_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Message_Chunk_Encryption::_Internal {
 public:
  using HasBits = decltype(::std::declval<Message_Chunk_Encryption>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_._has_bits_);
};

constexpr Message_Chunk_Encryption::ParseTableT_ Message_Chunk_Encryption::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk_Encryption>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated string enc_strings = 2;
      {::_pbi::TcParser::FastUR1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_.enc_strings_)}},
      // bool decrypted = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk_Encryption, _impl_.decrypted_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_.decrypted_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bool decrypted = 1;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_.decrypted_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // repeated string enc_strings = 2;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_.enc_strings_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
    }},
    // no aux_entries
    {{
      "\50\0\13\0\0\0\0\0"
      "AstroidMessages.Message.Chunk.Encryption"
      "enc_strings"
    }},
  };
}


inline constexpr Message_Chunk_Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        enc_strings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Encryption,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_.enc_strings_)>()
         }
        ,
        decrypted_{false} {}

template <typename>
constexpr Message_Chunk_Encryption::Message_Chunk_Encryption(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Message_Chunk_Encryption::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Message_Chunk_Encryption(arena);
}
constexpr auto Message_Chunk_Encryption::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Message_Chunk_Encryption), alignof(Message_Chunk_Encryption));
}
constexpr auto Message_Chunk_Encryption::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Message_Chunk_Encryption::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Message_Chunk_Encryption>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Message_Chunk_Encryption::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Message_Chunk_Encryption>(), &Message_Chunk_Encryption::ByteSizeLong,
              &Message_Chunk_Encryption::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Message_Chunk_Encryption, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[13],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct Message_Chunk_EncryptionGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr Message_Chunk_EncryptionGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Message_Chunk_Encryption_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Message_Chunk_Encryption::InternalGenerateClassData_(
            _default, &Message_Chunk_Encryption_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk_Encryption>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~Message_Chunk_EncryptionGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Message_Chunk_Encryption _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk_Encryption>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(Message_Chunk_EncryptionGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST Message_Chunk_EncryptionGlobalsTypeInternal Message_Chunk_Encryption_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Message_Chunk_Encryption_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Message_Chunk_Encryption_globals_.GetClassData();
#else
  return Message_Chunk_Encryption_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Mark::_Internal {
 public:
  using HasBits = decltype(::std::declval<Mark>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Mark, _impl_._has_bits_);
};

constexpr Mark::ParseTableT_ Mark::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Mark, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Mark>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool marked = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Mark, _impl_.marked_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Mark, _impl_.marked_)}},
      // string mid = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Mark, _impl_.mid_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string mid = 1;
      {PROTOBUF_FIELD_OFFSET(Mark, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool marked = 2;
      {PROTOBUF_FIELD_OFFSET(Mark, _impl_.marked_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\24\3\0\0\0\0\0\0"
      "AstroidMessages.Mark"
      "mid"
    }},
  };
}


inline constexpr Mark::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        marked_{false} {}

template <typename>
constexpr Mark::Mark(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Mark::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Mark(arena);
}
constexpr auto Mark::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Mark), alignof(Mark));
}
constexpr auto Mark::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Mark::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Mark>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Mark::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Mark>(), &Mark::ByteSizeLong,
              &Mark::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Mark, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[5],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct MarkGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr MarkGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Mark_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Mark::InternalGenerateClassData_(
            _default, &Mark_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Mark>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~MarkGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Mark _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Mark>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(MarkGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST MarkGlobalsTypeInternal Mark_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Mark_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Mark_globals_.GetClassData();
#else
  return Mark_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Info::_Internal {
 public:
  using HasBits = decltype(::std::declval<Info>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Info, _impl_._has_bits_);
};

constexpr Info::ParseTableT_ Info::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Info, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Info>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string txt = 4;
      {::_pbi::TcParser::FastUS1,
       {34, 1, 0,
        PROTOBUF_FIELD_OFFSET(Info, _impl_.txt_)}},
      // bool warning = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Info, _impl_.warning_), 2>(),
       {8, 2, 0,
        PROTOBUF_FIELD_OFFSET(Info, _impl_.warning_)}},
      // bool set = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Info, _impl_.set_), 3>(),
       {16, 3, 0,
        PROTOBUF_FIELD_OFFSET(Info, _impl_.set_)}},
      // string mid = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(Info, _impl_.mid_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bool warning = 1;
      {PROTOBUF_FIELD_OFFSET(Info, _impl_.warning_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool set = 2;
      {PROTOBUF_FIELD_OFFSET(Info, _impl_.set_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // string mid = 3;
      {PROTOBUF_FIELD_OFFSET(Info, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string txt = 4;
      {PROTOBUF_FIELD_OFFSET(Info, _impl_.txt_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\24\0\0\3\3\0\0\0"
      "AstroidMessages.Info"
      "mid"
      "txt"
    }},
  };
}


inline constexpr Info::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        txt_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        warning_{false},
        set_{false} {}

template <typename>
constexpr Info::Info(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Info::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Info(arena);
}
constexpr auto Info::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Info), alignof(Info));
}
constexpr auto Info::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Info::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Info>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Info::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Info>(), &Info::ByteSizeLong,
              &Info::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Info, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[9],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct InfoGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr InfoGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Info_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Info::InternalGenerateClassData_(
            _default, &Info_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Info>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~InfoGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Info _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Info>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(InfoGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST InfoGlobalsTypeInternal Info_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Info_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Info_globals_.GetClassData();
#else
  return Info_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Indent::_Internal {
 public:
  using HasBits = decltype(::std::declval<Indent>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Indent, _impl_._has_bits_);
};

constexpr Indent::ParseTableT_ Indent::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Indent, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Indent>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool indent = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Indent, _impl_.indent_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Indent, _impl_.indent_)}},
      // string bogus = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Indent, _impl_.bogus_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string bogus = 1;
      {PROTOBUF_FIELD_OFFSET(Indent, _impl_.bogus_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool indent = 2;
      {PROTOBUF_FIELD_OFFSET(Indent, _impl_.indent_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\26\5\0\0\0\0\0\0"
      "AstroidMessages.Indent"
      "bogus"
    }},
  };
}


inline constexpr Indent::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        bogus_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        indent_{false} {}

template <typename>
constexpr Indent::Indent(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Indent::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Indent(arena);
}
constexpr auto Indent::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Indent), alignof(Indent));
}
constexpr auto Indent::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Indent::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Indent>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Indent::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Indent>(), &Indent::ByteSizeLong,
              &Indent::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Indent, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct IndentGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr IndentGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Indent_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Indent::InternalGenerateClassData_(
            _default, &Indent_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Indent>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~IndentGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Indent _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Indent>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(IndentGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST IndentGlobalsTypeInternal Indent_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Indent_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Indent_globals_.GetClassData();
#else
  return Indent_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Hidden::_Internal {
 public:
  using HasBits = decltype(::std::declval<Hidden>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Hidden, _impl_._has_bits_);
};

constexpr Hidden::ParseTableT_ Hidden::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Hidden, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Hidden>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool hidden = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Hidden, _impl_.hidden_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Hidden, _impl_.hidden_)}},
      // string mid = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Hidden, _impl_.mid_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string mid = 1;
      {PROTOBUF_FIELD_OFFSET(Hidden, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool hidden = 2;
      {PROTOBUF_FIELD_OFFSET(Hidden, _impl_.hidden_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\26\3\0\0\0\0\0\0"
      "AstroidMessages.Hidden"
      "mid"
    }},
  };
}


inline constexpr Hidden::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        hidden_{false} {}

template <typename>
constexpr Hidden::Hidden(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Hidden::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Hidden(arena);
}
constexpr auto Hidden::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Hidden), alignof(Hidden));
}
constexpr auto Hidden::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Hidden::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Hidden>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Hidden::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Hidden>(), &Hidden::ByteSizeLong,
              &Hidden::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Hidden, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[6],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct HiddenGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr HiddenGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Hidden_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Hidden::InternalGenerateClassData_(
            _default, &Hidden_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Hidden>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~HiddenGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Hidden _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Hidden>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(HiddenGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST HiddenGlobalsTypeInternal Hidden_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Hidden_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Hidden_globals_.GetClassData();
#else
  return Hidden_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Focus::_Internal {
 public:
  using HasBits = decltype(::std::declval<Focus>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Focus, _impl_._has_bits_);
};

constexpr Focus::ParseTableT_ Focus::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Focus, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Focus>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string mid = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Focus, _impl_.mid_)}},
      // bool focus = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Focus, _impl_.focus_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Focus, _impl_.focus_)}},
      // int32 element = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Focus, _impl_.element_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(Focus, _impl_.element_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string mid = 1;
      {PROTOBUF_FIELD_OFFSET(Focus, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool focus = 2;
      {PROTOBUF_FIELD_OFFSET(Focus, _impl_.focus_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // int32 element = 3;
      {PROTOBUF_FIELD_OFFSET(Focus, _impl_.element_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
    }},
    // no aux_entries
    {{
      "\25\3\0\0\0\0\0\0"
      "AstroidMessages.Focus"
      "mid"
    }},
  };
}


inline constexpr Focus::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        focus_{false},
        element_{0} {}

template <typename>
constexpr Focus::Focus(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Focus::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Focus(arena);
}
constexpr auto Focus::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Focus), alignof(Focus));
}
constexpr auto Focus::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Focus::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Focus>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Focus::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Focus>(), &Focus::ByteSizeLong,
              &Focus::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Focus, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FocusGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FocusGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Focus_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Focus::InternalGenerateClassData_(
            _default, &Focus_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Focus>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FocusGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Focus _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Focus>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FocusGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FocusGlobalsTypeInternal Focus_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Focus_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Focus_globals_.GetClassData();
#else
  return Focus_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Debug::_Internal {
 public:
  using HasBits = decltype(::std::declval<Debug>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Debug, _impl_._has_bits_);
};

constexpr Debug::ParseTableT_ Debug::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Debug, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Debug>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string msg = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Debug, _impl_.msg_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string msg = 1;
      {PROTOBUF_FIELD_OFFSET(Debug, _impl_.msg_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\25\3\0\0\0\0\0\0"
      "AstroidMessages.Debug"
      "msg"
    }},
  };
}


inline constexpr Debug::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        msg_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Debug::Debug(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Debug::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Debug(arena);
}
constexpr auto Debug::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Debug), alignof(Debug));
}
constexpr auto Debug::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Debug::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Debug>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Debug::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Debug>(), &Debug::ByteSizeLong,
              &Debug::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Debug, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[7],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DebugGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DebugGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Debug_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Debug::InternalGenerateClassData_(
            _default, &Debug_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Debug>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DebugGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Debug _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Debug>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DebugGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DebugGlobalsTypeInternal Debug_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Debug_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Debug_globals_.GetClassData();
#else
  return Debug_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ClearMessage::_Internal {
 public:
  using HasBits = decltype(::std::declval<ClearMessage>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ClearMessage, _impl_._has_bits_);
};

constexpr ClearMessage::ParseTableT_ ClearMessage::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ClearMessage, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::ClearMessage>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool yes = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(ClearMessage, _impl_.yes_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(ClearMessage, _impl_.yes_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bool yes = 1;
      {PROTOBUF_FIELD_OFFSET(ClearMessage, _impl_.yes_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ClearMessage::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        yes_{false} {}

template <typename>
constexpr ClearMessage::ClearMessage(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ClearMessage::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ClearMessage(arena);
}
constexpr auto ClearMessage::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ClearMessage), alignof(ClearMessage));
}
constexpr auto ClearMessage::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ClearMessage::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ClearMessage>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ClearMessage::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ClearMessage>(), &ClearMessage::ByteSizeLong,
              &ClearMessage::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ClearMessage, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[20],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ClearMessageGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ClearMessageGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ClearMessage_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ClearMessage::InternalGenerateClassData_(
            _default, &ClearMessage_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ClearMessage>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ClearMessageGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ClearMessage _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ClearMessage>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ClearMessageGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ClearMessageGlobalsTypeInternal ClearMessage_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ClearMessage_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ClearMessage_globals_.GetClassData();
#else
  return ClearMessage_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AllowRemoteImages::_Internal {
 public:
  using HasBits = decltype(::std::declval<AllowRemoteImages>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_._has_bits_);
};

constexpr AllowRemoteImages::ParseTableT_ AllowRemoteImages::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::AllowRemoteImages>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bool allow = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(AllowRemoteImages, _impl_.allow_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_.allow_)}},
      // string bogus = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_.bogus_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string bogus = 1;
      {PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_.bogus_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool allow = 2;
      {PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_.allow_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\41\5\0\0\0\0\0\0"
      "AstroidMessages.AllowRemoteImages"
      "bogus"
    }},
  };
}


inline constexpr AllowRemoteImages::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        bogus_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        allow_{false} {}

template <typename>
constexpr AllowRemoteImages::AllowRemoteImages(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AllowRemoteImages::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AllowRemoteImages(arena);
}
constexpr auto AllowRemoteImages::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AllowRemoteImages), alignof(AllowRemoteImages));
}
constexpr auto AllowRemoteImages::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AllowRemoteImages::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AllowRemoteImages>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AllowRemoteImages::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AllowRemoteImages>(), &AllowRemoteImages::ByteSizeLong,
              &AllowRemoteImages::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AllowRemoteImages, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[3],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AllowRemoteImagesGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AllowRemoteImagesGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AllowRemoteImages_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AllowRemoteImages::InternalGenerateClassData_(
            _default, &AllowRemoteImages_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AllowRemoteImages>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AllowRemoteImagesGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AllowRemoteImages _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AllowRemoteImages>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AllowRemoteImagesGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AllowRemoteImagesGlobalsTypeInternal AllowRemoteImages_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AllowRemoteImages_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AllowRemoteImages_globals_.GetClassData();
#else
  return AllowRemoteImages_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Address::_Internal {
 public:
  using HasBits = decltype(::std::declval<Address>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Address, _impl_._has_bits_);
};

constexpr Address::ParseTableT_ Address::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Address, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Address>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string name = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Address, _impl_.name_)}},
      // string email = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(Address, _impl_.email_)}},
      // string full_address = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(Address, _impl_.full_address_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string name = 1;
      {PROTOBUF_FIELD_OFFSET(Address, _impl_.name_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string email = 2;
      {PROTOBUF_FIELD_OFFSET(Address, _impl_.email_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string full_address = 3;
      {PROTOBUF_FIELD_OFFSET(Address, _impl_.full_address_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\27\4\5\14\0\0\0\0"
      "AstroidMessages.Address"
      "name"
      "email"
      "full_address"
    }},
  };
}


inline constexpr Address::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        email_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        full_address_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Address::Address(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Address::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Address(arena);
}
constexpr auto Address::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Address), alignof(Address));
}
constexpr auto Address::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Address::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Address>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Address::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Address>(), &Address::ByteSizeLong,
              &Address::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Address, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[10],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AddressGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AddressGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Address_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Address::InternalGenerateClassData_(
            _default, &Address_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Address>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AddressGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Address _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Address>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AddressGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AddressGlobalsTypeInternal Address_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Address_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Address_globals_.GetClassData();
#else
  return Address_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class State_MessageState::_Internal {
 public:
  using HasBits = decltype(::std::declval<State_MessageState>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_._has_bits_);
};

constexpr State_MessageState::ParseTableT_ State_MessageState::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_._has_bits_),
      0, // no _extensions_
      6, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967246,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::State_MessageState>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // int32 level = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(State_MessageState, _impl_.level_), 1>(),
       {48, 1, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_.level_)}},
      // string mid = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_.mid_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string mid = 1;
      {PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_.mid_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated .AstroidMessages.State.MessageState.Element elements = 5;
      {PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_.elements_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // int32 level = 6;
      {PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_.level_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::State_MessageState_Element>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::State_MessageState_Element_globals_},
        #endif
    }},
    {{
      "\42\3\0\0\0\0\0\0"
      "AstroidMessages.State.MessageState"
      "mid"
    }},
  };
}


inline constexpr State_MessageState::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        level_{0},
        elements_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::State_MessageState,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.elements_)>()
         }
     {}

template <typename>
constexpr State_MessageState::State_MessageState(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL State_MessageState::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) State_MessageState(arena);
}
constexpr auto State_MessageState::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(State_MessageState), alignof(State_MessageState));
}
constexpr auto State_MessageState::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &State_MessageState::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<State_MessageState>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &State_MessageState::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<State_MessageState>(), &State_MessageState::ByteSizeLong,
              &State_MessageState::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(State_MessageState, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[17],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct State_MessageStateGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr State_MessageStateGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 State_MessageState_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(State_MessageState::InternalGenerateClassData_(
            _default, &State_MessageState_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<State_MessageState>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~State_MessageStateGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) State_MessageState _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<State_MessageState>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(State_MessageStateGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST State_MessageStateGlobalsTypeInternal State_MessageState_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* State_MessageState_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return State_MessageState_globals_.GetClassData();
#else
  return State_MessageState_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Message_Chunk::_Internal {
 public:
  using HasBits = decltype(::std::declval<Message_Chunk>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_._has_bits_);
};

constexpr Message_Chunk::ParseTableT_ Message_Chunk::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_._has_bits_),
      0, // no _extensions_
      22, 248,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4290772992,  // skipmap
      offsetof(ParseTableT_, field_entries),
      22,  // num_field_entries
      4,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // int32 id = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Message_Chunk, _impl_.id_), 11>(),
       {8, 11, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.id_)}},
      // bool viewable = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.viewable_), 12>(),
       {16, 12, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.viewable_)}},
      // bool preferred = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.preferred_), 13>(),
       {24, 13, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.preferred_)}},
      // repeated .AstroidMessages.Message.Chunk kids = 4;
      {::_pbi::TcParser::FastMtR1,
       {34, 0, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.kids_)}},
      // repeated .AstroidMessages.Message.Chunk siblings = 5;
      {::_pbi::TcParser::FastMtR1,
       {42, 1, 1,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.siblings_)}},
      // string mime_type = 6;
      {::_pbi::TcParser::FastUS1,
       {50, 2, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.mime_type_)}},
      // bool attachment = 7;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.attachment_), 14>(),
       {56, 14, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.attachment_)}},
      // bool is_encrypted = 8;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.is_encrypted_), 15>(),
       {64, 15, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.is_encrypted_)}},
      // bool is_signed = 9;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.is_signed_), 16>(),
       {72, 16, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.is_signed_)}},
      // string content = 10;
      {::_pbi::TcParser::FastUS1,
       {82, 3, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.content_)}},
      // bool sibling = 11;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.sibling_), 17>(),
       {88, 17, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.sibling_)}},
      // bool use = 12;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message_Chunk, _impl_.use_), 18>(),
       {96, 18, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.use_)}},
      // string sid = 13;
      {::_pbi::TcParser::FastUS1,
       {106, 4, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.sid_)}},
      // string filename = 14;
      {::_pbi::TcParser::FastUS1,
       {114, 5, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.filename_)}},
      // int32 size = 15;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Message_Chunk, _impl_.size_), 20>(),
       {120, 20, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.size_)}},
      // string human_size = 16;
      {::_pbi::TcParser::FastUS2,
       {386, 6, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.human_size_)}},
      // string thumbnail = 17;
      {::_pbi::TcParser::FastUS2,
       {394, 7, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.thumbnail_)}},
      // bool focusable = 18;
      {::_pbi::TcParser::FastV8S2,
       {400, 19, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.focusable_)}},
      // .AstroidMessages.Message.Chunk.Signature signature = 19;
      {::_pbi::TcParser::FastMtS2,
       {410, 9, 2,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.signature_)}},
      // .AstroidMessages.Message.Chunk.Encryption encryption = 20;
      {::_pbi::TcParser::FastMtS2,
       {418, 10, 3,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.encryption_)}},
      // int32 crypto_id = 21;
      {::_pbi::TcParser::FastV32S2,
       {424, 21, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.crypto_id_)}},
      // string cid = 22;
      {::_pbi::TcParser::FastUS2,
       {434, 8, 0,
        PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.cid_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // int32 id = 1;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.id_), _Internal::kHasBitsOffset + 11, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // bool viewable = 2;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.viewable_), _Internal::kHasBitsOffset + 12, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool preferred = 3;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.preferred_), _Internal::kHasBitsOffset + 13, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // repeated .AstroidMessages.Message.Chunk kids = 4;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.kids_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .AstroidMessages.Message.Chunk siblings = 5;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.siblings_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // string mime_type = 6;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.mime_type_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool attachment = 7;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.attachment_), _Internal::kHasBitsOffset + 14, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool is_encrypted = 8;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.is_encrypted_), _Internal::kHasBitsOffset + 15, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool is_signed = 9;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.is_signed_), _Internal::kHasBitsOffset + 16, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // string content = 10;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.content_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool sibling = 11;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.sibling_), _Internal::kHasBitsOffset + 17, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool use = 12;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.use_), _Internal::kHasBitsOffset + 18, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // string sid = 13;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.sid_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string filename = 14;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.filename_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // int32 size = 15;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.size_), _Internal::kHasBitsOffset + 20, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // string human_size = 16;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.human_size_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string thumbnail = 17;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.thumbnail_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool focusable = 18;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.focusable_), _Internal::kHasBitsOffset + 19, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // .AstroidMessages.Message.Chunk.Signature signature = 19;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.signature_), _Internal::kHasBitsOffset + 9, 2, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.Message.Chunk.Encryption encryption = 20;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.encryption_), _Internal::kHasBitsOffset + 10, 3, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // int32 crypto_id = 21;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.crypto_id_), _Internal::kHasBitsOffset + 21, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // string cid = 22;
      {PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.cid_), _Internal::kHasBitsOffset + 8, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk_Signature>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_Signature_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk_Encryption>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_Encryption_globals_},
        #endif
    }},
    {{
      "\35\0\0\0\0\0\11\0\0\0\7\0\0\3\10\0\12\11\0\0\0\0\3\0"
      "AstroidMessages.Message.Chunk"
      "mime_type"
      "content"
      "sid"
      "filename"
      "human_size"
      "thumbnail"
      "cid"
    }},
  };
}


inline constexpr Message_Chunk::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        kids_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.kids_)>()
         }
        ,
        siblings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.siblings_)>()
         }
        ,
        mime_type_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        content_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        sid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        filename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        human_size_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        thumbnail_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        cid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        signature_{nullptr},
        encryption_{nullptr},
        id_{0},
        viewable_{false},
        preferred_{false},
        attachment_{false},
        is_encrypted_{false},
        is_signed_{false},
        sibling_{false},
        use_{false},
        focusable_{false},
        size_{0},
        crypto_id_{0} {}

template <typename>
constexpr Message_Chunk::Message_Chunk(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Message_Chunk::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Message_Chunk(arena);
}
constexpr auto Message_Chunk::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Message_Chunk), alignof(Message_Chunk));
}
constexpr auto Message_Chunk::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Message_Chunk::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Message_Chunk>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Message_Chunk::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Message_Chunk>(), &Message_Chunk::ByteSizeLong,
              &Message_Chunk::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[14],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct Message_ChunkGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr Message_ChunkGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Message_Chunk_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Message_Chunk::InternalGenerateClassData_(
            _default, &Message_Chunk_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~Message_ChunkGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Message_Chunk _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Message_Chunk>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(Message_ChunkGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST Message_ChunkGlobalsTypeInternal Message_Chunk_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Message_Chunk_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Message_Chunk_globals_.GetClassData();
#else
  return Message_Chunk_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AddressList::_Internal {
 public:
  using HasBits = decltype(::std::declval<AddressList>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AddressList, _impl_._has_bits_);
};

constexpr AddressList::ParseTableT_ AddressList::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AddressList, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::AddressList>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .AstroidMessages.Address addresses = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AddressList, _impl_.addresses_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .AstroidMessages.Address addresses = 1;
      {PROTOBUF_FIELD_OFFSET(AddressList, _impl_.addresses_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Address>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Address_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr AddressList::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        addresses_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::AddressList,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::AddressList, _impl_.addresses_)>()
         }
     {}

template <typename>
constexpr AddressList::AddressList(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AddressList::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AddressList(arena);
}
constexpr auto AddressList::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AddressList), alignof(AddressList));
}
constexpr auto AddressList::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AddressList::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AddressList>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AddressList::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AddressList>(), &AddressList::ByteSizeLong,
              &AddressList::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AddressList, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[11],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AddressListGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AddressListGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AddressList_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AddressList::InternalGenerateClassData_(
            _default, &AddressList_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AddressList>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AddressListGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AddressList _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AddressList>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AddressListGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AddressListGlobalsTypeInternal AddressList_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AddressList_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AddressList_globals_.GetClassData();
#else
  return AddressList_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Ack::_Internal {
 public:
  using HasBits = decltype(::std::declval<Ack>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Ack, _impl_._has_bits_);
};

constexpr Ack::ParseTableT_ Ack::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Ack, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Ack>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // int32 id = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Ack, _impl_.id_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(Ack, _impl_.id_)}},
      // bool success = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Ack, _impl_.success_), 2>(),
       {16, 2, 0,
        PROTOBUF_FIELD_OFFSET(Ack, _impl_.success_)}},
      // .AstroidMessages.Focus focus = 3;
      {::_pbi::TcParser::FastMtS1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(Ack, _impl_.focus_)}},
    }}, {{
      65535, 65535
    }}, {{
      // int32 id = 1;
      {PROTOBUF_FIELD_OFFSET(Ack, _impl_.id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // bool success = 2;
      {PROTOBUF_FIELD_OFFSET(Ack, _impl_.success_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // .AstroidMessages.Focus focus = 3;
      {PROTOBUF_FIELD_OFFSET(Ack, _impl_.focus_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Focus>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Focus_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr Ack::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        focus_{nullptr},
        id_{0},
        success_{false} {}

template <typename>
constexpr Ack::Ack(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Ack::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Ack(arena);
}
constexpr auto Ack::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(Ack), alignof(Ack));
}
constexpr auto Ack::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Ack::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Ack>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Ack::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Ack>(), &Ack::ByteSizeLong,
              &Ack::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Ack, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AckGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AckGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Ack_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Ack::InternalGenerateClassData_(
            _default, &Ack_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Ack>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AckGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Ack _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Ack>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AckGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AckGlobalsTypeInternal Ack_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Ack_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Ack_globals_.GetClassData();
#else
  return Ack_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class State::_Internal {
 public:
  using HasBits = decltype(::std::declval<State>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(State, _impl_._has_bits_);
};

constexpr State::ParseTableT_ State::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(State, _impl_._has_bits_),
      0, // no _extensions_
      3, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967289,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::State>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .AstroidMessages.State.MessageState messages = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(State, _impl_.messages_)}},
      // bool edit_mode = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(State, _impl_.edit_mode_), 1>(),
       {24, 1, 0,
        PROTOBUF_FIELD_OFFSET(State, _impl_.edit_mode_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .AstroidMessages.State.MessageState messages = 2;
      {PROTOBUF_FIELD_OFFSET(State, _impl_.messages_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // bool edit_mode = 3;
      {PROTOBUF_FIELD_OFFSET(State, _impl_.edit_mode_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::State_MessageState>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::State_MessageState_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr State::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        messages_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::State,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_.messages_)>()
         }
        ,
        edit_mode_{false} {}

template <typename>
constexpr State::State(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL State::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) State(arena);
}
constexpr auto State::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(State), alignof(State));
}
constexpr auto State::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &State::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<State>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &State::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<State>(), &State::ByteSizeLong,
              &State::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(State, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[18],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StateGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StateGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 State_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(State::InternalGenerateClassData_(
            _default, &State_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<State>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StateGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) State _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<State>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StateGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StateGlobalsTypeInternal State_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* State_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return State_globals_.GetClassData();
#else
  return State_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Message::_Internal {
 public:
  using HasBits = decltype(::std::declval<Message>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Message, _impl_._has_bits_);
};

constexpr Message::ParseTableT_ Message::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Message, _impl_._has_bits_),
      0, // no _extensions_
      23, 248,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4286580864,  // skipmap
      offsetof(ParseTableT_, field_entries),
      21,  // num_field_entries
      8,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::Message>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string mid = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 3, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.mid_)}},
      // .AstroidMessages.Address sender = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 11, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.sender_)}},
      // .AstroidMessages.AddressList to = 3;
      {::_pbi::TcParser::FastMtS1,
       {26, 12, 1,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.to_)}},
      // .AstroidMessages.AddressList cc = 4;
      {::_pbi::TcParser::FastMtS1,
       {34, 13, 2,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.cc_)}},
      // .AstroidMessages.AddressList bcc = 5;
      {::_pbi::TcParser::FastMtS1,
       {42, 14, 3,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.bcc_)}},
      // string date_pretty = 6;
      {::_pbi::TcParser::FastUS1,
       {50, 4, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.date_pretty_)}},
      // string date_verbose = 7;
      {::_pbi::TcParser::FastUS1,
       {58, 5, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.date_verbose_)}},
      {::_pbi::TcParser::MiniParse, {}},
      // string subject = 9;
      {::_pbi::TcParser::FastUS1,
       {74, 6, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.subject_)}},
      // repeated string tags = 10;
      {::_pbi::TcParser::FastUR1,
       {82, 0, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.tags_)}},
      // string gravatar = 11;
      {::_pbi::TcParser::FastUS1,
       {90, 7, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.gravatar_)}},
      {::_pbi::TcParser::MiniParse, {}},
      // bool missing_content = 13;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message, _impl_.missing_content_), 18>(),
       {104, 18, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.missing_content_)}},
      // bool patch = 14;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Message, _impl_.patch_), 19>(),
       {112, 19, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.patch_)}},
      // int32 level = 15;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Message, _impl_.level_), 17>(),
       {120, 17, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.level_)}},
      // string in_reply_to = 16;
      {::_pbi::TcParser::FastUS2,
       {386, 8, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.in_reply_to_)}},
      // string preview = 17;
      {::_pbi::TcParser::FastUS2,
       {394, 9, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.preview_)}},
      // repeated .AstroidMessages.Message.Chunk mime_messages = 18;
      {::_pbi::TcParser::FastMtR2,
       {402, 1, 4,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.mime_messages_)}},
      // repeated .AstroidMessages.Message.Chunk attachments = 19;
      {::_pbi::TcParser::FastMtR2,
       {410, 2, 5,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.attachments_)}},
      // .AstroidMessages.Address reply_to = 20;
      {::_pbi::TcParser::FastMtS2,
       {418, 15, 6,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.reply_to_)}},
      // string tag_string = 21;
      {::_pbi::TcParser::FastUS2,
       {426, 10, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.tag_string_)}},
      // bool different_subject = 22;
      {::_pbi::TcParser::FastV8S2,
       {432, 20, 0,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.different_subject_)}},
      // .AstroidMessages.Message.Chunk root = 23;
      {::_pbi::TcParser::FastMtS2,
       {442, 16, 7,
        PROTOBUF_FIELD_OFFSET(Message, _impl_.root_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string mid = 1;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.mid_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // .AstroidMessages.Address sender = 2;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.sender_), _Internal::kHasBitsOffset + 11, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.AddressList to = 3;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.to_), _Internal::kHasBitsOffset + 12, 1, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.AddressList cc = 4;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.cc_), _Internal::kHasBitsOffset + 13, 2, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.AddressList bcc = 5;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.bcc_), _Internal::kHasBitsOffset + 14, 3, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // string date_pretty = 6;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.date_pretty_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string date_verbose = 7;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.date_verbose_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string subject = 9;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.subject_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated string tags = 10;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.tags_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
      // string gravatar = 11;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.gravatar_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool missing_content = 13;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.missing_content_), _Internal::kHasBitsOffset + 18, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool patch = 14;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.patch_), _Internal::kHasBitsOffset + 19, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // int32 level = 15;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.level_), _Internal::kHasBitsOffset + 17, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // string in_reply_to = 16;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.in_reply_to_), _Internal::kHasBitsOffset + 8, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string preview = 17;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.preview_), _Internal::kHasBitsOffset + 9, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated .AstroidMessages.Message.Chunk mime_messages = 18;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.mime_messages_), _Internal::kHasBitsOffset + 1, 4, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .AstroidMessages.Message.Chunk attachments = 19;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.attachments_), _Internal::kHasBitsOffset + 2, 5, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.Address reply_to = 20;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.reply_to_), _Internal::kHasBitsOffset + 15, 6, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // string tag_string = 21;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.tag_string_), _Internal::kHasBitsOffset + 10, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool different_subject = 22;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.different_subject_), _Internal::kHasBitsOffset + 20, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // .AstroidMessages.Message.Chunk root = 23;
      {PROTOBUF_FIELD_OFFSET(Message, _impl_.root_), _Internal::kHasBitsOffset + 16, 7, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Address>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Address_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::AddressList>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::AddressList_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::AddressList>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::AddressList_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::AddressList>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::AddressList_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Address>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Address_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message_Chunk>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_Chunk_globals_},
        #endif
    }},
    {{
      "\27\3\0\0\0\0\13\14\7\4\10\0\0\0\13\7\0\0\0\12\0\0\0\0"
      "AstroidMessages.Message"
      "mid"
      "date_pretty"
      "date_verbose"
      "subject"
      "tags"
      "gravatar"
      "in_reply_to"
      "preview"
      "tag_string"
    }},
  };
}


inline constexpr Message::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        tags_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.tags_)>()
         }
        ,
        mime_messages_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.mime_messages_)>()
         }
        ,
        attachments_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.attachments_)>()
         }
        ,
        mid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        date_pretty_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        date_verbose_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        subject_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        gravatar_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        in_reply_to_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        preview_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        tag_string_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        sender_{nullptr},
        to_{nullptr},
        cc_{nullptr},
        bcc_{nullptr},
        reply_to_{nullptr},
        root_{nullptr},
        level_{0},
        missing_content_{false},
        patch_{false},
        different_subject_{false} {}

template <typename>
constexpr Message::Message(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Message::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Message(arena);
}
constexpr auto Message::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Message), alignof(Message));
}
constexpr auto Message::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Message::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Message>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Message::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Message>(), &Message::ByteSizeLong,
              &Message::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Message, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[15],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct MessageGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr MessageGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Message_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Message::InternalGenerateClassData_(
            _default, &Message_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Message>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~MessageGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Message _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Message>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(MessageGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST MessageGlobalsTypeInternal Message_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Message_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Message_globals_.GetClassData();
#else
  return Message_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class UpdateMessage::_Internal {
 public:
  using HasBits = decltype(::std::declval<UpdateMessage>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_._has_bits_);
};

constexpr UpdateMessage::ParseTableT_ UpdateMessage::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::AstroidMessages::UpdateMessage>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // .AstroidMessages.UpdateMessage.Type type = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(UpdateMessage, _impl_.type_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.type_)}},
      // .AstroidMessages.Message m = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.m_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .AstroidMessages.Message m = 1;
      {PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.m_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .AstroidMessages.UpdateMessage.Type type = 2;
      {PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.type_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::AstroidMessages::Message>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::AstroidMessages::Message_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr UpdateMessage::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        m_{nullptr},
        type_{static_cast< ::AstroidMessages::UpdateMessage_Type >(0)} {}

template <typename>
constexpr UpdateMessage::UpdateMessage(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL UpdateMessage::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) UpdateMessage(arena);
}
constexpr auto UpdateMessage::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(UpdateMessage), alignof(UpdateMessage));
}
constexpr auto UpdateMessage::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &UpdateMessage::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<UpdateMessage>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &UpdateMessage::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<UpdateMessage>(), &UpdateMessage::ByteSizeLong,
              &UpdateMessage::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[19],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_messages_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct UpdateMessageGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr UpdateMessageGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 UpdateMessage_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(UpdateMessage::InternalGenerateClassData_(
            _default, &UpdateMessage_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<UpdateMessage>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~UpdateMessageGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) UpdateMessage _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<UpdateMessage>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(UpdateMessageGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST UpdateMessageGlobalsTypeInternal UpdateMessage_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* UpdateMessage_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return UpdateMessage_globals_.GetClassData();
#else
  return UpdateMessage_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace AstroidMessages
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_messages_2eproto[4];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_messages_2eproto = nullptr;
const ::uint32_t
    TableStruct_messages_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Focus, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Focus, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Focus, _impl_.focus_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Focus, _impl_.element_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Ack, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Ack, _impl_.id_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Ack, _impl_.success_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Ack, _impl_.focus_),
        1,
        2,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Indent, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Indent, _impl_.bogus_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Indent, _impl_.indent_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::AllowRemoteImages, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::AllowRemoteImages, _impl_.bogus_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::AllowRemoteImages, _impl_.allow_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_.direction_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_.type_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_.element_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Navigate, _impl_.focus_top_),
        1,
        2,
        0,
        3,
        4,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Mark, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Mark, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Mark, _impl_.marked_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Hidden, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Hidden, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Hidden, _impl_.hidden_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Debug, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Debug, _impl_.msg_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_._has_bits_),
        11, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.html_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.css_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.part_css_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.allowed_uris_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.use_stdout_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.use_syslog_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.disable_log_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.log_level_),
        1,
        2,
        3,
        0,
        5,
        6,
        7,
        4,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Info, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Info, _impl_.warning_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Info, _impl_.set_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Info, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Info, _impl_.txt_),
        2,
        3,
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Address, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Address, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Address, _impl_.email_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Address, _impl_.full_address_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::AddressList, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::AddressList, _impl_.addresses_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.verified_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.sign_strings_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.all_errors_),
        2,
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_.decrypted_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_.enc_strings_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_._has_bits_),
        25, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.id_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.sid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.mime_type_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.cid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.viewable_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.preferred_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.attachment_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.is_encrypted_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.is_signed_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.crypto_id_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.signature_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.encryption_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.sibling_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.use_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.focusable_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.content_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.filename_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.size_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.human_size_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.thumbnail_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.kids_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.siblings_),
        11,
        4,
        2,
        8,
        12,
        13,
        14,
        15,
        16,
        21,
        9,
        10,
        17,
        18,
        19,
        3,
        5,
        20,
        6,
        7,
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_._has_bits_),
        24, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.sender_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.to_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.cc_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.bcc_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.reply_to_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.date_pretty_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.date_verbose_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.subject_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.tags_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.tag_string_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.gravatar_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.missing_content_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.patch_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.different_subject_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.level_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.in_reply_to_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.preview_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.root_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.mime_messages_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.attachments_),
        3,
        11,
        12,
        13,
        14,
        15,
        4,
        5,
        6,
        0,
        10,
        7,
        18,
        19,
        20,
        17,
        8,
        9,
        16,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState_Element, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState_Element, _impl_.type_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState_Element, _impl_.id_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState_Element, _impl_.sid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState_Element, _impl_.focusable_),
        1,
        2,
        0,
        3,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.mid_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.elements_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.level_),
        0,
        2,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_.messages_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_.edit_mode_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::UpdateMessage, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::UpdateMessage, _impl_.m_),
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::UpdateMessage, _impl_.type_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::ClearMessage, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::AstroidMessages::ClearMessage, _impl_.yes_),
        0,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::AstroidMessages::Focus)},
        {9, sizeof(::AstroidMessages::Ack)},
        {18, sizeof(::AstroidMessages::Indent)},
        {25, sizeof(::AstroidMessages::AllowRemoteImages)},
        {32, sizeof(::AstroidMessages::Navigate)},
        {45, sizeof(::AstroidMessages::Mark)},
        {52, sizeof(::AstroidMessages::Hidden)},
        {59, sizeof(::AstroidMessages::Debug)},
        {64, sizeof(::AstroidMessages::Page)},
        {83, sizeof(::AstroidMessages::Info)},
        {94, sizeof(::AstroidMessages::Address)},
        {103, sizeof(::AstroidMessages::AddressList)},
        {108, sizeof(::AstroidMessages::Message_Chunk_Signature)},
        {117, sizeof(::AstroidMessages::Message_Chunk_Encryption)},
        {124, sizeof(::AstroidMessages::Message_Chunk)},
        {171, sizeof(::AstroidMessages::Message)},
        {216, sizeof(::AstroidMessages::State_MessageState_Element)},
        {227, sizeof(::AstroidMessages::State_MessageState)},
        {236, sizeof(::AstroidMessages::State)},
        {243, sizeof(::AstroidMessages::UpdateMessage)},
        {250, sizeof(::AstroidMessages::ClearMessage)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::AstroidMessages::Focus_globals_,
        &::AstroidMessages::Ack_globals_,
        &::AstroidMessages::Indent_globals_,
        &::AstroidMessages::AllowRemoteImages_globals_,
        &::AstroidMessages::Navigate_globals_,
        &::AstroidMessages::Mark_globals_,
        &::AstroidMessages::Hidden_globals_,
        &::AstroidMessages::Debug_globals_,
        &::AstroidMessages::Page_globals_,
        &::AstroidMessages::Info_globals_,
        &::AstroidMessages::Address_globals_,
        &::AstroidMessages::AddressList_globals_,
        &::AstroidMessages::Message_Chunk_Signature_globals_,
        &::AstroidMessages::Message_Chunk_Encryption_globals_,
        &::AstroidMessages::Message_Chunk_globals_,
        &::AstroidMessages::Message_globals_,
        &::AstroidMessages::State_MessageState_Element_globals_,
        &::AstroidMessages::State_MessageState_globals_,
        &::AstroidMessages::State_globals_,
        &::AstroidMessages::UpdateMessage_globals_,
        &::AstroidMessages::ClearMessage_globals_,
};
const char descriptor_table_protodef_messages_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\016messages.proto\022\017AstroidMessages\"4\n\005Foc"
    "us\022\013\n\003mid\030\001 \001(\t\022\r\n\005focus\030\002 \001(\010\022\017\n\007elemen"
    "t\030\003 \001(\005\"I\n\003Ack\022\n\n\002id\030\001 \001(\005\022\017\n\007success\030\002 "
    "\001(\010\022%\n\005focus\030\003 \001(\0132\026.AstroidMessages.Foc"
    "us\"\'\n\006Indent\022\r\n\005bogus\030\001 \001(\t\022\016\n\006indent\030\002 "
    "\001(\010\"1\n\021AllowRemoteImages\022\r\n\005bogus\030\001 \001(\t\022"
    "\r\n\005allow\030\002 \001(\010\"\324\002\n\010Navigate\0226\n\tdirection"
    "\030\001 \001(\0162#.AstroidMessages.Navigate.Direct"
    "ion\022,\n\004type\030\002 \001(\0162\036.AstroidMessages.Navi"
    "gate.Type\022\013\n\003mid\030\003 \001(\t\022\017\n\007element\030\004 \001(\005\022"
    "\021\n\tfocus_top\030\005 \001(\010\"5\n\tDirection\022\010\n\004None\020"
    "\000\022\014\n\010Specific\020\001\022\006\n\002Up\020\002\022\010\n\004Down\020\003\"z\n\004Typ"
    "e\022\021\n\rVisualElement\020\000\022\n\n\006Visual\020\001\022\r\n\tVisu"
    "alBig\020\002\022\016\n\nVisualPage\020\003\022\013\n\007Element\020\004\022\013\n\007"
    "Message\020\005\022\r\n\tFocusView\020\006\022\013\n\007Extreme\020\007\"#\n"
    "\004Mark\022\013\n\003mid\030\001 \001(\t\022\016\n\006marked\030\002 \001(\010\"%\n\006Hi"
    "dden\022\013\n\003mid\030\001 \001(\t\022\016\n\006hidden\030\002 \001(\010\"\024\n\005Deb"
    "ug\022\013\n\003msg\030\001 \001(\t\"\231\001\n\004Page\022\014\n\004html\030\001 \001(\t\022\013"
    "\n\003css\030\002 \001(\t\022\020\n\010part_css\030\003 \001(\t\022\024\n\014allowed"
    "_uris\030\004 \003(\t\022\022\n\nuse_stdout\030\005 \001(\010\022\022\n\nuse_s"
    "yslog\030\006 \001(\010\022\023\n\013disable_log\030\007 \001(\010\022\021\n\tlog_"
    "level\030\010 \001(\t\">\n\004Info\022\017\n\007warning\030\001 \001(\010\022\013\n\003"
    "set\030\002 \001(\010\022\013\n\003mid\030\003 \001(\t\022\013\n\003txt\030\004 \001(\t\"<\n\007A"
    "ddress\022\014\n\004name\030\001 \001(\t\022\r\n\005email\030\002 \001(\t\022\024\n\014f"
    "ull_address\030\003 \001(\t\":\n\013AddressList\022+\n\taddr"
    "esses\030\001 \003(\0132\030.AstroidMessages.Address\"\211\n"
    "\n\007Message\022\013\n\003mid\030\001 \001(\t\022(\n\006sender\030\002 \001(\0132\030"
    ".AstroidMessages.Address\022(\n\002to\030\003 \001(\0132\034.A"
    "stroidMessages.AddressList\022(\n\002cc\030\004 \001(\0132\034"
    ".AstroidMessages.AddressList\022)\n\003bcc\030\005 \001("
    "\0132\034.AstroidMessages.AddressList\022*\n\010reply"
    "_to\030\024 \001(\0132\030.AstroidMessages.Address\022\023\n\013d"
    "ate_pretty\030\006 \001(\t\022\024\n\014date_verbose\030\007 \001(\t\022\017"
    "\n\007subject\030\t \001(\t\022\014\n\004tags\030\n \003(\t\022\022\n\ntag_str"
    "ing\030\025 \001(\t\022\020\n\010gravatar\030\013 \001(\t\022\027\n\017missing_c"
    "ontent\030\r \001(\010\022\r\n\005patch\030\016 \001(\010\022\031\n\021different"
    "_subject\030\026 \001(\010\022\r\n\005level\030\017 \001(\005\022\023\n\013in_repl"
    "y_to\030\020 \001(\t\022\017\n\007preview\030\021 \001(\t\022,\n\004root\030\027 \001("
    "\0132\036.AstroidMessages.Message.Chunk\0225\n\rmim"
    "e_messages\030\022 \003(\0132\036.AstroidMessages.Messa"
    "ge.Chunk\0223\n\013attachments\030\023 \003(\0132\036.AstroidM"
    "essages.Message.Chunk\032\231\005\n\005Chunk\022\n\n\002id\030\001 "
    "\001(\005\022\013\n\003sid\030\r \001(\t\022\021\n\tmime_type\030\006 \001(\t\022\013\n\003c"
    "id\030\026 \001(\t\022\020\n\010viewable\030\002 \001(\010\022\021\n\tpreferred\030"
    "\003 \001(\010\022\022\n\nattachment\030\007 \001(\010\022\024\n\014is_encrypte"
    "d\030\010 \001(\010\022\021\n\tis_signed\030\t \001(\010\022\021\n\tcrypto_id\030"
    "\025 \001(\005\022;\n\tsignature\030\023 \001(\0132(.AstroidMessag"
    "es.Message.Chunk.Signature\022=\n\nencryption"
    "\030\024 \001(\0132).AstroidMessages.Message.Chunk.E"
    "ncryption\022\017\n\007sibling\030\013 \001(\010\022\013\n\003use\030\014 \001(\010\022"
    "\021\n\tfocusable\030\022 \001(\010\022\017\n\007content\030\n \001(\t\022\020\n\010f"
    "ilename\030\016 \001(\t\022\014\n\004size\030\017 \001(\005\022\022\n\nhuman_siz"
    "e\030\020 \001(\t\022\021\n\tthumbnail\030\021 \001(\t\022,\n\004kids\030\004 \003(\013"
    "2\036.AstroidMessages.Message.Chunk\0220\n\010sibl"
    "ings\030\005 \003(\0132\036.AstroidMessages.Message.Chu"
    "nk\032G\n\tSignature\022\020\n\010verified\030\001 \001(\010\022\024\n\014sig"
    "n_strings\030\002 \003(\t\022\022\n\nall_errors\030\003 \003(\t\0324\n\nE"
    "ncryption\022\021\n\tdecrypted\030\001 \001(\010\022\023\n\013enc_stri"
    "ngs\030\002 \003(\t\"\220\003\n\005State\0225\n\010messages\030\002 \003(\0132#."
    "AstroidMessages.State.MessageState\022\021\n\ted"
    "it_mode\030\003 \001(\010\032\274\002\n\014MessageState\022\013\n\003mid\030\001 "
    "\001(\t\022=\n\010elements\030\005 \003(\0132+.AstroidMessages."
    "State.MessageState.Element\022\r\n\005level\030\006 \001("
    "\005\032\320\001\n\007Element\022>\n\004type\030\001 \001(\01620.AstroidMes"
    "sages.State.MessageState.Element.Type\022\n\n"
    "\002id\030\002 \001(\005\022\013\n\003sid\030\003 \001(\t\022\021\n\tfocusable\030\004 \001("
    "\010\"Y\n\004Type\022\t\n\005Empty\020\000\022\013\n\007Address\020\001\022\010\n\004Par"
    "t\020\002\022\016\n\nAttachment\020\003\022\017\n\013MimeMessage\020\004\022\016\n\n"
    "Encryption\020\005\"\213\001\n\rUpdateMessage\022#\n\001m\030\001 \001("
    "\0132\030.AstroidMessages.Message\0221\n\004type\030\002 \001("
    "\0162#.AstroidMessages.UpdateMessage.Type\"\""
    "\n\004Type\022\010\n\004Tags\020\000\022\020\n\014VisibleParts\020\001\"\033\n\014Cl"
    "earMessage\022\013\n\003yes\030\001 \001(\010b\006proto3"
};
static ::absl::once_flag descriptor_table_messages_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_messages_2eproto = {
    false,
    false,
    2911,
    descriptor_table_protodef_messages_2eproto,
    "messages.proto",
    &descriptor_table_messages_2eproto_once,
    nullptr,
    0,
    21,
    schemas,
    file_message_globals,
    TableStruct_messages_2eproto::offsets,
    file_level_enum_descriptors_messages_2eproto,
    file_level_service_descriptors_messages_2eproto,
};
namespace AstroidMessages {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Navigate_Direction_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_messages_2eproto);
  return file_level_enum_descriptors_messages_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t Navigate_Direction_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Navigate_Type_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_messages_2eproto);
  return file_level_enum_descriptors_messages_2eproto[1];
}
PROTOBUF_CONSTINIT const uint32_t Navigate_Type_internal_data_[] = {
    524288u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
State_MessageState_Element_Type_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_messages_2eproto);
  return file_level_enum_descriptors_messages_2eproto[2];
}
PROTOBUF_CONSTINIT const uint32_t State_MessageState_Element_Type_internal_data_[] = {
    393216u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
UpdateMessage_Type_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_messages_2eproto);
  return file_level_enum_descriptors_messages_2eproto[3];
}
PROTOBUF_CONSTINIT const uint32_t UpdateMessage_Type_internal_data_[] = {
    131072u, 0u, };
// ===================================================================

Focus::Focus(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Focus_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Focus)
}
PROTOBUF_NDEBUG_INLINE Focus::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Focus& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_) {}

Focus::Focus(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Focus& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Focus_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Focus* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, focus_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, focus_),
           offsetof(Impl_, element_) -
               offsetof(Impl_, focus_) +
               sizeof(Impl_::element_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Focus)
}
PROTOBUF_NDEBUG_INLINE Focus::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena) {}

inline void Focus::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, focus_),
           0,
           offsetof(Impl_, element_) -
               offsetof(Impl_, focus_) +
               sizeof(Impl_::element_));
}
Focus::~Focus() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Focus)
  SharedDtor(*this);
}
inline void Focus::SharedDtor(MessageLite& self) {
  Focus& this_ = static_cast<Focus&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Focus_class_data_ =
        Focus::InternalGenerateClassData_(Focus_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Focus::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Focus_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Focus_class_data_.tc_table);
  return Focus_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Focus::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Focus_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Focus_globals_));
  return Focus_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Focus::ParseTableT_
    Focus::_table_ =
        Focus::InternalGenerateParseTable_(Focus_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Focus::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Focus)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.mid_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    ::memset(&_impl_.focus_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.element_) -
        reinterpret_cast<char*>(&_impl_.focus_)) + sizeof(_impl_.element_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Focus::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Focus& this_ = static_cast<const Focus&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Focus::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Focus& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Focus)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string mid = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Focus.mid");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bool focus = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_focus() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_focus(), target);
    }
  }

  // int32 element = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_element() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<3>(
              stream, this_._internal_element(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Focus)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Focus::ByteSizeLong(const MessageLite& base) {
  const Focus& this_ = static_cast<const Focus&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Focus::ByteSizeLong() const {
  const Focus& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Focus)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string mid = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // bool focus = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_focus() != 0) {
        total_size += 2;
      }
    }
    // int32 element = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_element() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_element());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Focus::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Focus*>(&to_msg);
  auto& from = static_cast<const Focus&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Focus)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_focus() != 0) {
        _this->_impl_.focus_ = from._impl_.focus_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_element() != 0) {
        _this->_impl_.element_ = from._impl_.element_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Focus::CopyFrom(const Focus& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Focus)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Focus::InternalSwap(Focus* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Focus, _impl_.element_)
      + sizeof(Focus::_impl_.element_)
      - PROTOBUF_FIELD_OFFSET(Focus, _impl_.focus_)>(
          reinterpret_cast<char*>(&_impl_.focus_),
          reinterpret_cast<char*>(&other->_impl_.focus_));
}

::google::protobuf::Metadata Focus::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Ack::Ack(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Ack_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Ack)
}
PROTOBUF_NDEBUG_INLINE Ack::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Ack& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

Ack::Ack(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Ack& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Ack_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Ack* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.focus_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.focus_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, id_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, id_),
           offsetof(Impl_, success_) -
               offsetof(Impl_, id_) +
               sizeof(Impl_::success_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Ack)
}
PROTOBUF_NDEBUG_INLINE Ack::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void Ack::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, focus_),
           0,
           offsetof(Impl_, success_) -
               offsetof(Impl_, focus_) +
               sizeof(Impl_::success_));
}
Ack::~Ack() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Ack)
  SharedDtor(*this);
}
inline void Ack::SharedDtor(MessageLite& self) {
  Ack& this_ = static_cast<Ack&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.focus_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Ack_class_data_ =
        Ack::InternalGenerateClassData_(Ack_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Ack::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Ack_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Ack_class_data_.tc_table);
  return Ack_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Ack::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Ack_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Ack_globals_));
  return Ack_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Ack::ParseTableT_
    Ack::_table_ =
        Ack::InternalGenerateParseTable_(Ack_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Ack::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Ack)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(_impl_.focus_ != nullptr);
    _impl_.focus_->Clear();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    ::memset(&_impl_.id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.success_) -
        reinterpret_cast<char*>(&_impl_.id_)) + sizeof(_impl_.success_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Ack::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Ack& this_ = static_cast<const Ack&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Ack::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Ack& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Ack)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // int32 id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<1>(
              stream, this_._internal_id(), target);
    }
  }

  // bool success = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_success() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_success(), target);
    }
  }

  // .AstroidMessages.Focus focus = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.focus_, this_._impl_.focus_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Ack)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Ack::ByteSizeLong(const MessageLite& base) {
  const Ack& this_ = static_cast<const Ack&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Ack::ByteSizeLong() const {
  const Ack& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Ack)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // .AstroidMessages.Focus focus = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.focus_);
    }
    // int32 id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_id());
      }
    }
    // bool success = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_success() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Ack::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Ack*>(&to_msg);
  auto& from = static_cast<const Ack&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Ack)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.focus_ != nullptr);
      if (_this->_impl_.focus_ == nullptr) {
        _this->_impl_.focus_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.focus_);
      } else {
        _this->_impl_.focus_->MergeFrom(*from._impl_.focus_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_id() != 0) {
        _this->_impl_.id_ = from._impl_.id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_success() != 0) {
        _this->_impl_.success_ = from._impl_.success_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Ack::CopyFrom(const Ack& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Ack)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Ack::InternalSwap(Ack* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Ack, _impl_.success_)
      + sizeof(Ack::_impl_.success_)
      - PROTOBUF_FIELD_OFFSET(Ack, _impl_.focus_)>(
          reinterpret_cast<char*>(&_impl_.focus_),
          reinterpret_cast<char*>(&other->_impl_.focus_));
}

::google::protobuf::Metadata Ack::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Indent::Indent(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Indent_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Indent)
}
PROTOBUF_NDEBUG_INLINE Indent::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Indent& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        bogus_(arena, from.bogus_) {}

Indent::Indent(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Indent& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Indent_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Indent* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.indent_ = from._impl_.indent_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Indent)
}
PROTOBUF_NDEBUG_INLINE Indent::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        bogus_(arena) {}

inline void Indent::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.indent_ = {};
}
Indent::~Indent() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Indent)
  SharedDtor(*this);
}
inline void Indent::SharedDtor(MessageLite& self) {
  Indent& this_ = static_cast<Indent&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.bogus_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Indent_class_data_ =
        Indent::InternalGenerateClassData_(Indent_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Indent::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Indent_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Indent_class_data_.tc_table);
  return Indent_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Indent::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Indent_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Indent_globals_));
  return Indent_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Indent::ParseTableT_
    Indent::_table_ =
        Indent::InternalGenerateParseTable_(Indent_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Indent::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Indent)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.bogus_.ClearNonDefaultToEmpty();
  }
  _impl_.indent_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Indent::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Indent& this_ = static_cast<const Indent&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Indent::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Indent& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Indent)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string bogus = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_bogus().empty()) {
      const ::std::string& _s = this_._internal_bogus();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Indent.bogus");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bool indent = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_indent() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_indent(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Indent)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Indent::ByteSizeLong(const MessageLite& base) {
  const Indent& this_ = static_cast<const Indent&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Indent::ByteSizeLong() const {
  const Indent& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Indent)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string bogus = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_bogus().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_bogus());
      }
    }
    // bool indent = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_indent() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Indent::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Indent*>(&to_msg);
  auto& from = static_cast<const Indent&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Indent)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_bogus().empty()) {
        _this->_internal_set_bogus(from._internal_bogus());
      } else {
        if (_this->_impl_.bogus_.IsDefault()) {
          _this->_internal_set_bogus("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_indent() != 0) {
        _this->_impl_.indent_ = from._impl_.indent_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Indent::CopyFrom(const Indent& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Indent)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Indent::InternalSwap(Indent* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.bogus_, &other->_impl_.bogus_, arena);
  swap(_impl_.indent_, other->_impl_.indent_);
}

::google::protobuf::Metadata Indent::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

AllowRemoteImages::AllowRemoteImages(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AllowRemoteImages_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.AllowRemoteImages)
}
PROTOBUF_NDEBUG_INLINE AllowRemoteImages::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::AllowRemoteImages& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        bogus_(arena, from.bogus_) {}

AllowRemoteImages::AllowRemoteImages(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AllowRemoteImages& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AllowRemoteImages_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AllowRemoteImages* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.allow_ = from._impl_.allow_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.AllowRemoteImages)
}
PROTOBUF_NDEBUG_INLINE AllowRemoteImages::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        bogus_(arena) {}

inline void AllowRemoteImages::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.allow_ = {};
}
AllowRemoteImages::~AllowRemoteImages() {
  // @@protoc_insertion_point(destructor:AstroidMessages.AllowRemoteImages)
  SharedDtor(*this);
}
inline void AllowRemoteImages::SharedDtor(MessageLite& self) {
  AllowRemoteImages& this_ = static_cast<AllowRemoteImages&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.bogus_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AllowRemoteImages_class_data_ =
        AllowRemoteImages::InternalGenerateClassData_(AllowRemoteImages_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AllowRemoteImages::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AllowRemoteImages_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AllowRemoteImages_class_data_.tc_table);
  return AllowRemoteImages_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AllowRemoteImages::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AllowRemoteImages_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AllowRemoteImages_globals_));
  return AllowRemoteImages_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AllowRemoteImages::ParseTableT_
    AllowRemoteImages::_table_ =
        AllowRemoteImages::InternalGenerateParseTable_(AllowRemoteImages_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AllowRemoteImages::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.AllowRemoteImages)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.bogus_.ClearNonDefaultToEmpty();
  }
  _impl_.allow_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AllowRemoteImages::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AllowRemoteImages& this_ = static_cast<const AllowRemoteImages&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AllowRemoteImages::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AllowRemoteImages& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.AllowRemoteImages)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string bogus = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_bogus().empty()) {
      const ::std::string& _s = this_._internal_bogus();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.AllowRemoteImages.bogus");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bool allow = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_allow() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_allow(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.AllowRemoteImages)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AllowRemoteImages::ByteSizeLong(const MessageLite& base) {
  const AllowRemoteImages& this_ = static_cast<const AllowRemoteImages&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AllowRemoteImages::ByteSizeLong() const {
  const AllowRemoteImages& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.AllowRemoteImages)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string bogus = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_bogus().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_bogus());
      }
    }
    // bool allow = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_allow() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AllowRemoteImages::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AllowRemoteImages*>(&to_msg);
  auto& from = static_cast<const AllowRemoteImages&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.AllowRemoteImages)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_bogus().empty()) {
        _this->_internal_set_bogus(from._internal_bogus());
      } else {
        if (_this->_impl_.bogus_.IsDefault()) {
          _this->_internal_set_bogus("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_allow() != 0) {
        _this->_impl_.allow_ = from._impl_.allow_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AllowRemoteImages::CopyFrom(const AllowRemoteImages& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.AllowRemoteImages)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AllowRemoteImages::InternalSwap(AllowRemoteImages* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.bogus_, &other->_impl_.bogus_, arena);
  swap(_impl_.allow_, other->_impl_.allow_);
}

::google::protobuf::Metadata AllowRemoteImages::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Navigate::Navigate(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Navigate_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Navigate)
}
PROTOBUF_NDEBUG_INLINE Navigate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Navigate& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_) {}

Navigate::Navigate(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Navigate& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Navigate_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Navigate* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, direction_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, direction_),
           offsetof(Impl_, focus_top_) -
               offsetof(Impl_, direction_) +
               sizeof(Impl_::focus_top_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Navigate)
}
PROTOBUF_NDEBUG_INLINE Navigate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena) {}

inline void Navigate::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, direction_),
           0,
           offsetof(Impl_, focus_top_) -
               offsetof(Impl_, direction_) +
               sizeof(Impl_::focus_top_));
}
Navigate::~Navigate() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Navigate)
  SharedDtor(*this);
}
inline void Navigate::SharedDtor(MessageLite& self) {
  Navigate& this_ = static_cast<Navigate&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Navigate_class_data_ =
        Navigate::InternalGenerateClassData_(Navigate_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Navigate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Navigate_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Navigate_class_data_.tc_table);
  return Navigate_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Navigate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Navigate_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Navigate_globals_));
  return Navigate_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Navigate::ParseTableT_
    Navigate::_table_ =
        Navigate::InternalGenerateParseTable_(Navigate_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Navigate::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Navigate)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.mid_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000001eU)) {
    ::memset(&_impl_.direction_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.focus_top_) -
        reinterpret_cast<char*>(&_impl_.direction_)) + sizeof(_impl_.focus_top_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Navigate::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Navigate& this_ = static_cast<const Navigate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Navigate::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Navigate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Navigate)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .AstroidMessages.Navigate.Direction direction = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_direction() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_direction(), target);
    }
  }

  // .AstroidMessages.Navigate.Type type = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_type() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          2, this_._internal_type(), target);
    }
  }

  // string mid = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Navigate.mid");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // int32 element = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_element() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<4>(
              stream, this_._internal_element(), target);
    }
  }

  // bool focus_top = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_focus_top() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          5, this_._internal_focus_top(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Navigate)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Navigate::ByteSizeLong(const MessageLite& base) {
  const Navigate& this_ = static_cast<const Navigate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Navigate::ByteSizeLong() const {
  const Navigate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Navigate)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // string mid = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // .AstroidMessages.Navigate.Direction direction = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_direction() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_direction());
      }
    }
    // .AstroidMessages.Navigate.Type type = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_type() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_type());
      }
    }
    // int32 element = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_element() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_element());
      }
    }
    // bool focus_top = 5;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_focus_top() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Navigate::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Navigate*>(&to_msg);
  auto& from = static_cast<const Navigate&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Navigate)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_direction() != 0) {
        _this->_impl_.direction_ = from._impl_.direction_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_type() != 0) {
        _this->_impl_.type_ = from._impl_.type_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_element() != 0) {
        _this->_impl_.element_ = from._impl_.element_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_focus_top() != 0) {
        _this->_impl_.focus_top_ = from._impl_.focus_top_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Navigate::CopyFrom(const Navigate& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Navigate)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Navigate::InternalSwap(Navigate* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Navigate, _impl_.focus_top_)
      + sizeof(Navigate::_impl_.focus_top_)
      - PROTOBUF_FIELD_OFFSET(Navigate, _impl_.direction_)>(
          reinterpret_cast<char*>(&_impl_.direction_),
          reinterpret_cast<char*>(&other->_impl_.direction_));
}

::google::protobuf::Metadata Navigate::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Mark::Mark(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Mark_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Mark)
}
PROTOBUF_NDEBUG_INLINE Mark::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Mark& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_) {}

Mark::Mark(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Mark& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Mark_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Mark* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.marked_ = from._impl_.marked_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Mark)
}
PROTOBUF_NDEBUG_INLINE Mark::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena) {}

inline void Mark::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.marked_ = {};
}
Mark::~Mark() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Mark)
  SharedDtor(*this);
}
inline void Mark::SharedDtor(MessageLite& self) {
  Mark& this_ = static_cast<Mark&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Mark_class_data_ =
        Mark::InternalGenerateClassData_(Mark_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Mark::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Mark_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Mark_class_data_.tc_table);
  return Mark_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Mark::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Mark_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Mark_globals_));
  return Mark_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Mark::ParseTableT_
    Mark::_table_ =
        Mark::InternalGenerateParseTable_(Mark_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Mark::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Mark)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.mid_.ClearNonDefaultToEmpty();
  }
  _impl_.marked_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Mark::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Mark& this_ = static_cast<const Mark&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Mark::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Mark& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Mark)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string mid = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Mark.mid");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bool marked = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_marked() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_marked(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Mark)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Mark::ByteSizeLong(const MessageLite& base) {
  const Mark& this_ = static_cast<const Mark&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Mark::ByteSizeLong() const {
  const Mark& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Mark)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string mid = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // bool marked = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_marked() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Mark::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Mark*>(&to_msg);
  auto& from = static_cast<const Mark&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Mark)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_marked() != 0) {
        _this->_impl_.marked_ = from._impl_.marked_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Mark::CopyFrom(const Mark& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Mark)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Mark::InternalSwap(Mark* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  swap(_impl_.marked_, other->_impl_.marked_);
}

::google::protobuf::Metadata Mark::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Hidden::Hidden(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Hidden_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Hidden)
}
PROTOBUF_NDEBUG_INLINE Hidden::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Hidden& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_) {}

Hidden::Hidden(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Hidden& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Hidden_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Hidden* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.hidden_ = from._impl_.hidden_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Hidden)
}
PROTOBUF_NDEBUG_INLINE Hidden::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena) {}

inline void Hidden::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.hidden_ = {};
}
Hidden::~Hidden() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Hidden)
  SharedDtor(*this);
}
inline void Hidden::SharedDtor(MessageLite& self) {
  Hidden& this_ = static_cast<Hidden&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Hidden_class_data_ =
        Hidden::InternalGenerateClassData_(Hidden_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Hidden::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Hidden_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Hidden_class_data_.tc_table);
  return Hidden_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Hidden::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Hidden_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Hidden_globals_));
  return Hidden_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Hidden::ParseTableT_
    Hidden::_table_ =
        Hidden::InternalGenerateParseTable_(Hidden_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Hidden::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Hidden)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.mid_.ClearNonDefaultToEmpty();
  }
  _impl_.hidden_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Hidden::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Hidden& this_ = static_cast<const Hidden&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Hidden::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Hidden& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Hidden)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string mid = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Hidden.mid");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bool hidden = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_hidden() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_hidden(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Hidden)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Hidden::ByteSizeLong(const MessageLite& base) {
  const Hidden& this_ = static_cast<const Hidden&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Hidden::ByteSizeLong() const {
  const Hidden& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Hidden)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string mid = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // bool hidden = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_hidden() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Hidden::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Hidden*>(&to_msg);
  auto& from = static_cast<const Hidden&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Hidden)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_hidden() != 0) {
        _this->_impl_.hidden_ = from._impl_.hidden_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Hidden::CopyFrom(const Hidden& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Hidden)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Hidden::InternalSwap(Hidden* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  swap(_impl_.hidden_, other->_impl_.hidden_);
}

::google::protobuf::Metadata Hidden::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Debug::Debug(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Debug_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Debug)
}
PROTOBUF_NDEBUG_INLINE Debug::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Debug& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        msg_(arena, from.msg_) {}

Debug::Debug(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Debug& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Debug_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Debug* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Debug)
}
PROTOBUF_NDEBUG_INLINE Debug::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        msg_(arena) {}

inline void Debug::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Debug::~Debug() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Debug)
  SharedDtor(*this);
}
inline void Debug::SharedDtor(MessageLite& self) {
  Debug& this_ = static_cast<Debug&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.msg_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Debug_class_data_ =
        Debug::InternalGenerateClassData_(Debug_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Debug::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Debug_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Debug_class_data_.tc_table);
  return Debug_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Debug::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Debug_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Debug_globals_));
  return Debug_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Debug::ParseTableT_
    Debug::_table_ =
        Debug::InternalGenerateParseTable_(Debug_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Debug::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Debug)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.msg_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Debug::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Debug& this_ = static_cast<const Debug&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Debug::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Debug& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Debug)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string msg = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_msg().empty()) {
      const ::std::string& _s = this_._internal_msg();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Debug.msg");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Debug)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Debug::ByteSizeLong(const MessageLite& base) {
  const Debug& this_ = static_cast<const Debug&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Debug::ByteSizeLong() const {
  const Debug& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Debug)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // string msg = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_msg().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_msg());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Debug::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Debug*>(&to_msg);
  auto& from = static_cast<const Debug&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Debug)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_msg().empty()) {
      _this->_internal_set_msg(from._internal_msg());
    } else {
      if (_this->_impl_.msg_.IsDefault()) {
        _this->_internal_set_msg("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Debug::CopyFrom(const Debug& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Debug)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Debug::InternalSwap(Debug* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.msg_, &other->_impl_.msg_, arena);
}

::google::protobuf::Metadata Debug::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Page::Page(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Page_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Page)
}
PROTOBUF_NDEBUG_INLINE Page::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Page& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        allowed_uris_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Page,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.allowed_uris_)>()
          , from.allowed_uris_
        }
        ,
        html_(arena, from.html_),
        css_(arena, from.css_),
        part_css_(arena, from.part_css_),
        log_level_(arena, from.log_level_) {}

Page::Page(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Page& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Page_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Page* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, use_stdout_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, use_stdout_),
           offsetof(Impl_, disable_log_) -
               offsetof(Impl_, use_stdout_) +
               sizeof(Impl_::disable_log_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Page)
}
PROTOBUF_NDEBUG_INLINE Page::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        allowed_uris_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Page,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Page, _impl_.allowed_uris_)>()
         }
        ,
        html_(arena),
        css_(arena),
        part_css_(arena),
        log_level_(arena) {}

inline void Page::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, use_stdout_),
           0,
           offsetof(Impl_, disable_log_) -
               offsetof(Impl_, use_stdout_) +
               sizeof(Impl_::disable_log_));
}
Page::~Page() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Page)
  SharedDtor(*this);
}
inline void Page::SharedDtor(MessageLite& self) {
  Page& this_ = static_cast<Page&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.html_.Destroy();
  this_._impl_.css_.Destroy();
  this_._impl_.part_css_.Destroy();
  this_._impl_.log_level_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Page_class_data_ =
        Page::InternalGenerateClassData_(Page_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Page::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Page_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Page_class_data_.tc_table);
  return Page_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Page::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Page_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Page_globals_));
  return Page_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Page::ParseTableT_
    Page::_table_ =
        Page::InternalGenerateParseTable_(Page_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Page::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Page)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.allowed_uris_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.html_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.css_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.part_css_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.log_level_.ClearNonDefaultToEmpty();
    }
  }
  ::memset(&_impl_.use_stdout_, 0, static_cast<::size_t>(
      reinterpret_cast<char*>(&_impl_.disable_log_) -
      reinterpret_cast<char*>(&_impl_.use_stdout_)) + sizeof(_impl_.disable_log_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Page::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Page& this_ = static_cast<const Page&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Page::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Page& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Page)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string html = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_html().empty()) {
      const ::std::string& _s = this_._internal_html();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Page.html");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string css = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_css().empty()) {
      const ::std::string& _s = this_._internal_css();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Page.css");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string part_css = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_part_css().empty()) {
      const ::std::string& _s = this_._internal_part_css();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Page.part_css");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // repeated string allowed_uris = 4;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_allowed_uris_size(); i < n; ++i) {
      const auto& s = this_._internal_allowed_uris().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Page.allowed_uris");
      target = stream->WriteString(4, s, target);
    }
  }

  // bool use_stdout = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_use_stdout() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          5, this_._internal_use_stdout(), target);
    }
  }

  // bool use_syslog = 6;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (this_._internal_use_syslog() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          6, this_._internal_use_syslog(), target);
    }
  }

  // bool disable_log = 7;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    if (this_._internal_disable_log() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          7, this_._internal_disable_log(), target);
    }
  }

  // string log_level = 8;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (!this_._internal_log_level().empty()) {
      const ::std::string& _s = this_._internal_log_level();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Page.log_level");
      target = stream->WriteStringMaybeAliased(8, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Page)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Page::ByteSizeLong(const MessageLite& base) {
  const Page& this_ = static_cast<const Page&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Page::ByteSizeLong() const {
  const Page& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Page)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated string allowed_uris = 4;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_allowed_uris().size());
      for (int i = 0, n = this_._internal_allowed_uris().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_allowed_uris().Get(i));
      }
    }
    // string html = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_html().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_html());
      }
    }
    // string css = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_css().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_css());
      }
    }
    // string part_css = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_part_css().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_part_css());
      }
    }
    // string log_level = 8;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!this_._internal_log_level().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_log_level());
      }
    }
    // bool use_stdout = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_use_stdout() != 0) {
        total_size += 2;
      }
    }
    // bool use_syslog = 6;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (this_._internal_use_syslog() != 0) {
        total_size += 2;
      }
    }
    // bool disable_log = 7;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (this_._internal_disable_log() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Page::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Page*>(&to_msg);
  auto& from = static_cast<const Page&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Page)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_allowed_uris()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_allowed_uris());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_html().empty()) {
        _this->_internal_set_html(from._internal_html());
      } else {
        if (_this->_impl_.html_.IsDefault()) {
          _this->_internal_set_html("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_css().empty()) {
        _this->_internal_set_css(from._internal_css());
      } else {
        if (_this->_impl_.css_.IsDefault()) {
          _this->_internal_set_css("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_part_css().empty()) {
        _this->_internal_set_part_css(from._internal_part_css());
      } else {
        if (_this->_impl_.part_css_.IsDefault()) {
          _this->_internal_set_part_css("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!from._internal_log_level().empty()) {
        _this->_internal_set_log_level(from._internal_log_level());
      } else {
        if (_this->_impl_.log_level_.IsDefault()) {
          _this->_internal_set_log_level("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_use_stdout() != 0) {
        _this->_impl_.use_stdout_ = from._impl_.use_stdout_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (from._internal_use_syslog() != 0) {
        _this->_impl_.use_syslog_ = from._impl_.use_syslog_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (from._internal_disable_log() != 0) {
        _this->_impl_.disable_log_ = from._impl_.disable_log_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Page::CopyFrom(const Page& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Page)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Page::InternalSwap(Page* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.allowed_uris_.InternalSwap(&other->_impl_.allowed_uris_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.html_, &other->_impl_.html_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.css_, &other->_impl_.css_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.part_css_, &other->_impl_.part_css_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.log_level_, &other->_impl_.log_level_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Page, _impl_.disable_log_)
      + sizeof(Page::_impl_.disable_log_)
      - PROTOBUF_FIELD_OFFSET(Page, _impl_.use_stdout_)>(
          reinterpret_cast<char*>(&_impl_.use_stdout_),
          reinterpret_cast<char*>(&other->_impl_.use_stdout_));
}

::google::protobuf::Metadata Page::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Info::Info(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Info_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Info)
}
PROTOBUF_NDEBUG_INLINE Info::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Info& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_),
        txt_(arena, from.txt_) {}

Info::Info(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Info& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Info_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Info* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, warning_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, warning_),
           offsetof(Impl_, set_) -
               offsetof(Impl_, warning_) +
               sizeof(Impl_::set_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Info)
}
PROTOBUF_NDEBUG_INLINE Info::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena),
        txt_(arena) {}

inline void Info::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, warning_),
           0,
           offsetof(Impl_, set_) -
               offsetof(Impl_, warning_) +
               sizeof(Impl_::set_));
}
Info::~Info() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Info)
  SharedDtor(*this);
}
inline void Info::SharedDtor(MessageLite& self) {
  Info& this_ = static_cast<Info&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.txt_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Info_class_data_ =
        Info::InternalGenerateClassData_(Info_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Info::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Info_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Info_class_data_.tc_table);
  return Info_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Info::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Info_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Info_globals_));
  return Info_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Info::ParseTableT_
    Info::_table_ =
        Info::InternalGenerateParseTable_(Info_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Info::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Info)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.mid_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.txt_.ClearNonDefaultToEmpty();
    }
  }
  ::memset(&_impl_.warning_, 0, static_cast<::size_t>(
      reinterpret_cast<char*>(&_impl_.set_) -
      reinterpret_cast<char*>(&_impl_.warning_)) + sizeof(_impl_.set_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Info::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Info& this_ = static_cast<const Info&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Info::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Info& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Info)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bool warning = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_warning() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          1, this_._internal_warning(), target);
    }
  }

  // bool set = 2;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_set() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_set(), target);
    }
  }

  // string mid = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Info.mid");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // string txt = 4;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_txt().empty()) {
      const ::std::string& _s = this_._internal_txt();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Info.txt");
      target = stream->WriteStringMaybeAliased(4, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Info)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Info::ByteSizeLong(const MessageLite& base) {
  const Info& this_ = static_cast<const Info&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Info::ByteSizeLong() const {
  const Info& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Info)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // string mid = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // string txt = 4;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_txt().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_txt());
      }
    }
    // bool warning = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_warning() != 0) {
        total_size += 2;
      }
    }
    // bool set = 2;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_set() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Info::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Info*>(&to_msg);
  auto& from = static_cast<const Info&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Info)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_txt().empty()) {
        _this->_internal_set_txt(from._internal_txt());
      } else {
        if (_this->_impl_.txt_.IsDefault()) {
          _this->_internal_set_txt("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_warning() != 0) {
        _this->_impl_.warning_ = from._impl_.warning_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_set() != 0) {
        _this->_impl_.set_ = from._impl_.set_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Info::CopyFrom(const Info& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Info)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Info::InternalSwap(Info* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.txt_, &other->_impl_.txt_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Info, _impl_.set_)
      + sizeof(Info::_impl_.set_)
      - PROTOBUF_FIELD_OFFSET(Info, _impl_.warning_)>(
          reinterpret_cast<char*>(&_impl_.warning_),
          reinterpret_cast<char*>(&other->_impl_.warning_));
}

::google::protobuf::Metadata Info::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Address::Address(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Address_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Address)
}
PROTOBUF_NDEBUG_INLINE Address::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Address& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        name_(arena, from.name_),
        email_(arena, from.email_),
        full_address_(arena, from.full_address_) {}

Address::Address(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Address& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Address_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Address* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Address)
}
PROTOBUF_NDEBUG_INLINE Address::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        name_(arena),
        email_(arena),
        full_address_(arena) {}

inline void Address::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Address::~Address() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Address)
  SharedDtor(*this);
}
inline void Address::SharedDtor(MessageLite& self) {
  Address& this_ = static_cast<Address&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.name_.Destroy();
  this_._impl_.email_.Destroy();
  this_._impl_.full_address_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Address_class_data_ =
        Address::InternalGenerateClassData_(Address_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Address::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Address_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Address_class_data_.tc_table);
  return Address_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Address::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Address_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Address_globals_));
  return Address_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Address::ParseTableT_
    Address::_table_ =
        Address::InternalGenerateParseTable_(Address_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Address::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Address)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.email_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.full_address_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Address::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Address& this_ = static_cast<const Address&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Address::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Address& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Address)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_name().empty()) {
      const ::std::string& _s = this_._internal_name();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Address.name");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string email = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_email().empty()) {
      const ::std::string& _s = this_._internal_email();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Address.email");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string full_address = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_full_address().empty()) {
      const ::std::string& _s = this_._internal_full_address();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Address.full_address");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Address)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Address::ByteSizeLong(const MessageLite& base) {
  const Address& this_ = static_cast<const Address&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Address::ByteSizeLong() const {
  const Address& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Address)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_name().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_name());
      }
    }
    // string email = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_email().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_email());
      }
    }
    // string full_address = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_full_address().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_full_address());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Address::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Address*>(&to_msg);
  auto& from = static_cast<const Address&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Address)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_name().empty()) {
        _this->_internal_set_name(from._internal_name());
      } else {
        if (_this->_impl_.name_.IsDefault()) {
          _this->_internal_set_name("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_email().empty()) {
        _this->_internal_set_email(from._internal_email());
      } else {
        if (_this->_impl_.email_.IsDefault()) {
          _this->_internal_set_email("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_full_address().empty()) {
        _this->_internal_set_full_address(from._internal_full_address());
      } else {
        if (_this->_impl_.full_address_.IsDefault()) {
          _this->_internal_set_full_address("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Address::CopyFrom(const Address& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Address)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Address::InternalSwap(Address* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.email_, &other->_impl_.email_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.full_address_, &other->_impl_.full_address_, arena);
}

::google::protobuf::Metadata Address::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

AddressList::AddressList(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AddressList_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.AddressList)
}
PROTOBUF_NDEBUG_INLINE AddressList::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::AddressList& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        addresses_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::AddressList,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::AddressList, _impl_.addresses_)>()
          , from.addresses_
        }
     {}

AddressList::AddressList(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AddressList& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AddressList_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AddressList* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.AddressList)
}
PROTOBUF_NDEBUG_INLINE AddressList::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        addresses_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::AddressList,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::AddressList, _impl_.addresses_)>()
         }
     {}

inline void AddressList::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
AddressList::~AddressList() {
  // @@protoc_insertion_point(destructor:AstroidMessages.AddressList)
  SharedDtor(*this);
}
inline void AddressList::SharedDtor(MessageLite& self) {
  AddressList& this_ = static_cast<AddressList&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AddressList_class_data_ =
        AddressList::InternalGenerateClassData_(AddressList_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AddressList::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AddressList_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AddressList_class_data_.tc_table);
  return AddressList_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AddressList::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AddressList_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AddressList_globals_));
  return AddressList_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AddressList::ParseTableT_
    AddressList::_table_ =
        AddressList::InternalGenerateParseTable_(AddressList_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AddressList::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.AddressList)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.addresses_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AddressList::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AddressList& this_ = static_cast<const AddressList&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AddressList::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AddressList& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.AddressList)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .AstroidMessages.Address addresses = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_addresses_size());
         i < n; i++) {
      const auto& repfield = this_._internal_addresses().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.AddressList)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AddressList::ByteSizeLong(const MessageLite& base) {
  const AddressList& this_ = static_cast<const AddressList&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AddressList::ByteSizeLong() const {
  const AddressList& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.AddressList)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .AstroidMessages.Address addresses = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_addresses_size();
      for (const auto& msg : this_._internal_addresses()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AddressList::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AddressList*>(&to_msg);
  auto& from = static_cast<const AddressList&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.AddressList)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_addresses()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_addresses());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AddressList::CopyFrom(const AddressList& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.AddressList)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AddressList::InternalSwap(AddressList* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.addresses_.InternalSwap(&other->_impl_.addresses_);
}

::google::protobuf::Metadata AddressList::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Message_Chunk_Signature::Message_Chunk_Signature(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_Signature_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Message.Chunk.Signature)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk_Signature::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Message_Chunk_Signature& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        sign_strings_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message_Chunk_Signature,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.sign_strings_)>()
          , from.sign_strings_
        }
        ,
        all_errors_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message_Chunk_Signature,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.all_errors_)>()
          , from.all_errors_
        }
     {}

Message_Chunk_Signature::Message_Chunk_Signature(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Message_Chunk_Signature& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_Signature_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Message_Chunk_Signature* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.verified_ = from._impl_.verified_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Message.Chunk.Signature)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk_Signature::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        sign_strings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Signature,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.sign_strings_)>()
         }
        ,
        all_errors_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Signature,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Signature, _impl_.all_errors_)>()
         }
     {}

inline void Message_Chunk_Signature::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.verified_ = {};
}
Message_Chunk_Signature::~Message_Chunk_Signature() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Message.Chunk.Signature)
  SharedDtor(*this);
}
inline void Message_Chunk_Signature::SharedDtor(MessageLite& self) {
  Message_Chunk_Signature& this_ = static_cast<Message_Chunk_Signature&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Message_Chunk_Signature_class_data_ =
        Message_Chunk_Signature::InternalGenerateClassData_(Message_Chunk_Signature_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk_Signature::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_Signature_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Message_Chunk_Signature_class_data_.tc_table);
  return Message_Chunk_Signature_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk_Signature::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_Signature_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Message_Chunk_Signature_globals_));
  return Message_Chunk_Signature_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Message_Chunk_Signature::ParseTableT_
    Message_Chunk_Signature::_table_ =
        Message_Chunk_Signature::InternalGenerateParseTable_(Message_Chunk_Signature_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Message_Chunk_Signature::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Message.Chunk.Signature)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.sign_strings_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.all_errors_.Clear();
    }
  }
  _impl_.verified_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Message_Chunk_Signature::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Message_Chunk_Signature& this_ = static_cast<const Message_Chunk_Signature&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Message_Chunk_Signature::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Message_Chunk_Signature& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Message.Chunk.Signature)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bool verified = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_verified() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          1, this_._internal_verified(), target);
    }
  }

  // repeated string sign_strings = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_sign_strings_size(); i < n; ++i) {
      const auto& s = this_._internal_sign_strings().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.Signature.sign_strings");
      target = stream->WriteString(2, s, target);
    }
  }

  // repeated string all_errors = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (int i = 0, n = this_._internal_all_errors_size(); i < n; ++i) {
      const auto& s = this_._internal_all_errors().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.Signature.all_errors");
      target = stream->WriteString(3, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Message.Chunk.Signature)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Message_Chunk_Signature::ByteSizeLong(const MessageLite& base) {
  const Message_Chunk_Signature& this_ = static_cast<const Message_Chunk_Signature&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Message_Chunk_Signature::ByteSizeLong() const {
  const Message_Chunk_Signature& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Message.Chunk.Signature)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated string sign_strings = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_sign_strings().size());
      for (int i = 0, n = this_._internal_sign_strings().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_sign_strings().Get(i));
      }
    }
    // repeated string all_errors = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_all_errors().size());
      for (int i = 0, n = this_._internal_all_errors().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_all_errors().Get(i));
      }
    }
    // bool verified = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_verified() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Message_Chunk_Signature::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Message_Chunk_Signature*>(&to_msg);
  auto& from = static_cast<const Message_Chunk_Signature&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Message.Chunk.Signature)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_sign_strings()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_sign_strings());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_all_errors()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_all_errors());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_verified() != 0) {
        _this->_impl_.verified_ = from._impl_.verified_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Message_Chunk_Signature::CopyFrom(const Message_Chunk_Signature& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Message.Chunk.Signature)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Message_Chunk_Signature::InternalSwap(Message_Chunk_Signature* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.sign_strings_.InternalSwap(&other->_impl_.sign_strings_);
  _impl_.all_errors_.InternalSwap(&other->_impl_.all_errors_);
  swap(_impl_.verified_, other->_impl_.verified_);
}

::google::protobuf::Metadata Message_Chunk_Signature::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Message_Chunk_Encryption::Message_Chunk_Encryption(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_Encryption_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Message.Chunk.Encryption)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk_Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Message_Chunk_Encryption& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        enc_strings_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message_Chunk_Encryption,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_.enc_strings_)>()
          , from.enc_strings_
        }
     {}

Message_Chunk_Encryption::Message_Chunk_Encryption(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Message_Chunk_Encryption& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_Encryption_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Message_Chunk_Encryption* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.decrypted_ = from._impl_.decrypted_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Message.Chunk.Encryption)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk_Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        enc_strings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk_Encryption,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk_Encryption, _impl_.enc_strings_)>()
         }
     {}

inline void Message_Chunk_Encryption::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.decrypted_ = {};
}
Message_Chunk_Encryption::~Message_Chunk_Encryption() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Message.Chunk.Encryption)
  SharedDtor(*this);
}
inline void Message_Chunk_Encryption::SharedDtor(MessageLite& self) {
  Message_Chunk_Encryption& this_ = static_cast<Message_Chunk_Encryption&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Message_Chunk_Encryption_class_data_ =
        Message_Chunk_Encryption::InternalGenerateClassData_(Message_Chunk_Encryption_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk_Encryption::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_Encryption_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Message_Chunk_Encryption_class_data_.tc_table);
  return Message_Chunk_Encryption_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk_Encryption::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_Encryption_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Message_Chunk_Encryption_globals_));
  return Message_Chunk_Encryption_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Message_Chunk_Encryption::ParseTableT_
    Message_Chunk_Encryption::_table_ =
        Message_Chunk_Encryption::InternalGenerateParseTable_(Message_Chunk_Encryption_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Message_Chunk_Encryption::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Message.Chunk.Encryption)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.enc_strings_.Clear();
  }
  _impl_.decrypted_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Message_Chunk_Encryption::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Message_Chunk_Encryption& this_ = static_cast<const Message_Chunk_Encryption&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Message_Chunk_Encryption::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Message_Chunk_Encryption& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Message.Chunk.Encryption)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bool decrypted = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_decrypted() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          1, this_._internal_decrypted(), target);
    }
  }

  // repeated string enc_strings = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_enc_strings_size(); i < n; ++i) {
      const auto& s = this_._internal_enc_strings().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.Encryption.enc_strings");
      target = stream->WriteString(2, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Message.Chunk.Encryption)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Message_Chunk_Encryption::ByteSizeLong(const MessageLite& base) {
  const Message_Chunk_Encryption& this_ = static_cast<const Message_Chunk_Encryption&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Message_Chunk_Encryption::ByteSizeLong() const {
  const Message_Chunk_Encryption& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Message.Chunk.Encryption)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated string enc_strings = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_enc_strings().size());
      for (int i = 0, n = this_._internal_enc_strings().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_enc_strings().Get(i));
      }
    }
    // bool decrypted = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_decrypted() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Message_Chunk_Encryption::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Message_Chunk_Encryption*>(&to_msg);
  auto& from = static_cast<const Message_Chunk_Encryption&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Message.Chunk.Encryption)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_enc_strings()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_enc_strings());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_decrypted() != 0) {
        _this->_impl_.decrypted_ = from._impl_.decrypted_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Message_Chunk_Encryption::CopyFrom(const Message_Chunk_Encryption& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Message.Chunk.Encryption)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Message_Chunk_Encryption::InternalSwap(Message_Chunk_Encryption* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.enc_strings_.InternalSwap(&other->_impl_.enc_strings_);
  swap(_impl_.decrypted_, other->_impl_.decrypted_);
}

::google::protobuf::Metadata Message_Chunk_Encryption::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Message_Chunk::Message_Chunk(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Message.Chunk)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Message_Chunk& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        kids_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message_Chunk,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.kids_)>()
          , from.kids_
        }
        ,
        siblings_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message_Chunk,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.siblings_)>()
          , from.siblings_
        }
        ,
        mime_type_(arena, from.mime_type_),
        content_(arena, from.content_),
        sid_(arena, from.sid_),
        filename_(arena, from.filename_),
        human_size_(arena, from.human_size_),
        thumbnail_(arena, from.thumbnail_),
        cid_(arena, from.cid_) {}

Message_Chunk::Message_Chunk(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Message_Chunk& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_Chunk_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Message_Chunk* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.signature_ = (CheckHasBit(cached_has_bits, 0x00000200U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.signature_)
                : nullptr;
  _impl_.encryption_ = (CheckHasBit(cached_has_bits, 0x00000400U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.encryption_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, id_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, id_),
           offsetof(Impl_, crypto_id_) -
               offsetof(Impl_, id_) +
               sizeof(Impl_::crypto_id_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Message.Chunk)
}
PROTOBUF_NDEBUG_INLINE Message_Chunk::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        kids_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.kids_)>()
         }
        ,
        siblings_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message_Chunk,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message_Chunk, _impl_.siblings_)>()
         }
        ,
        mime_type_(arena),
        content_(arena),
        sid_(arena),
        filename_(arena),
        human_size_(arena),
        thumbnail_(arena),
        cid_(arena) {}

inline void Message_Chunk::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, signature_),
           0,
           offsetof(Impl_, crypto_id_) -
               offsetof(Impl_, signature_) +
               sizeof(Impl_::crypto_id_));
}
Message_Chunk::~Message_Chunk() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Message.Chunk)
  SharedDtor(*this);
}
inline void Message_Chunk::SharedDtor(MessageLite& self) {
  Message_Chunk& this_ = static_cast<Message_Chunk&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mime_type_.Destroy();
  this_._impl_.content_.Destroy();
  this_._impl_.sid_.Destroy();
  this_._impl_.filename_.Destroy();
  this_._impl_.human_size_.Destroy();
  this_._impl_.thumbnail_.Destroy();
  this_._impl_.cid_.Destroy();
  delete this_._impl_.signature_;
  delete this_._impl_.encryption_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Message_Chunk_class_data_ =
        Message_Chunk::InternalGenerateClassData_(Message_Chunk_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Message_Chunk_class_data_.tc_table);
  return Message_Chunk_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message_Chunk::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_Chunk_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Message_Chunk_globals_));
  return Message_Chunk_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Message_Chunk::ParseTableT_
    Message_Chunk::_table_ =
        Message_Chunk::InternalGenerateParseTable_(Message_Chunk_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Message_Chunk::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Message.Chunk)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.kids_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.siblings_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.mime_type_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.content_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.sid_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _impl_.filename_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _impl_.human_size_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _impl_.thumbnail_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000700U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      _impl_.cid_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      ABSL_DCHECK(_impl_.signature_ != nullptr);
      _impl_.signature_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      ABSL_DCHECK(_impl_.encryption_ != nullptr);
      _impl_.encryption_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000f800U)) {
    ::memset(&_impl_.id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.is_encrypted_) -
        reinterpret_cast<char*>(&_impl_.id_)) + sizeof(_impl_.is_encrypted_));
  }
  if (BatchCheckHasBit(cached_has_bits, 0x003f0000U)) {
    ::memset(&_impl_.is_signed_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.crypto_id_) -
        reinterpret_cast<char*>(&_impl_.is_signed_)) + sizeof(_impl_.crypto_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Message_Chunk::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Message_Chunk& this_ = static_cast<const Message_Chunk&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Message_Chunk::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Message_Chunk& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Message.Chunk)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // int32 id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000800U)) {
    if (this_._internal_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<1>(
              stream, this_._internal_id(), target);
    }
  }

  // bool viewable = 2;
  if (CheckHasBit(cached_has_bits, 0x00001000U)) {
    if (this_._internal_viewable() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_viewable(), target);
    }
  }

  // bool preferred = 3;
  if (CheckHasBit(cached_has_bits, 0x00002000U)) {
    if (this_._internal_preferred() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          3, this_._internal_preferred(), target);
    }
  }

  // repeated .AstroidMessages.Message.Chunk kids = 4;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_kids_size());
         i < n; i++) {
      const auto& repfield = this_._internal_kids().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              4, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .AstroidMessages.Message.Chunk siblings = 5;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_siblings_size());
         i < n; i++) {
      const auto& repfield = this_._internal_siblings().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              5, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // string mime_type = 6;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_mime_type().empty()) {
      const ::std::string& _s = this_._internal_mime_type();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.mime_type");
      target = stream->WriteStringMaybeAliased(6, _s, target);
    }
  }

  // bool attachment = 7;
  if (CheckHasBit(cached_has_bits, 0x00004000U)) {
    if (this_._internal_attachment() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          7, this_._internal_attachment(), target);
    }
  }

  // bool is_encrypted = 8;
  if (CheckHasBit(cached_has_bits, 0x00008000U)) {
    if (this_._internal_is_encrypted() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          8, this_._internal_is_encrypted(), target);
    }
  }

  // bool is_signed = 9;
  if (CheckHasBit(cached_has_bits, 0x00010000U)) {
    if (this_._internal_is_signed() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          9, this_._internal_is_signed(), target);
    }
  }

  // string content = 10;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_content().empty()) {
      const ::std::string& _s = this_._internal_content();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.content");
      target = stream->WriteStringMaybeAliased(10, _s, target);
    }
  }

  // bool sibling = 11;
  if (CheckHasBit(cached_has_bits, 0x00020000U)) {
    if (this_._internal_sibling() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          11, this_._internal_sibling(), target);
    }
  }

  // bool use = 12;
  if (CheckHasBit(cached_has_bits, 0x00040000U)) {
    if (this_._internal_use() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          12, this_._internal_use(), target);
    }
  }

  // string sid = 13;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (!this_._internal_sid().empty()) {
      const ::std::string& _s = this_._internal_sid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.sid");
      target = stream->WriteStringMaybeAliased(13, _s, target);
    }
  }

  // string filename = 14;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (!this_._internal_filename().empty()) {
      const ::std::string& _s = this_._internal_filename();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.filename");
      target = stream->WriteStringMaybeAliased(14, _s, target);
    }
  }

  // int32 size = 15;
  if (CheckHasBit(cached_has_bits, 0x00100000U)) {
    if (this_._internal_size() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<15>(
              stream, this_._internal_size(), target);
    }
  }

  // string human_size = 16;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (!this_._internal_human_size().empty()) {
      const ::std::string& _s = this_._internal_human_size();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.human_size");
      target = stream->WriteStringMaybeAliased(16, _s, target);
    }
  }

  // string thumbnail = 17;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    if (!this_._internal_thumbnail().empty()) {
      const ::std::string& _s = this_._internal_thumbnail();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.thumbnail");
      target = stream->WriteStringMaybeAliased(17, _s, target);
    }
  }

  // bool focusable = 18;
  if (CheckHasBit(cached_has_bits, 0x00080000U)) {
    if (this_._internal_focusable() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          18, this_._internal_focusable(), target);
    }
  }

  // .AstroidMessages.Message.Chunk.Signature signature = 19;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        19, *this_._impl_.signature_, this_._impl_.signature_->GetCachedSize(), target,
        stream);
  }

  // .AstroidMessages.Message.Chunk.Encryption encryption = 20;
  if (CheckHasBit(cached_has_bits, 0x00000400U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        20, *this_._impl_.encryption_, this_._impl_.encryption_->GetCachedSize(), target,
        stream);
  }

  // int32 crypto_id = 21;
  if (CheckHasBit(cached_has_bits, 0x00200000U)) {
    if (this_._internal_crypto_id() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteInt32ToArray(
          21, this_._internal_crypto_id(), target);
    }
  }

  // string cid = 22;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    if (!this_._internal_cid().empty()) {
      const ::std::string& _s = this_._internal_cid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.Chunk.cid");
      target = stream->WriteStringMaybeAliased(22, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Message.Chunk)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Message_Chunk::ByteSizeLong(const MessageLite& base) {
  const Message_Chunk& this_ = static_cast<const Message_Chunk&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Message_Chunk::ByteSizeLong() const {
  const Message_Chunk& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Message.Chunk)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated .AstroidMessages.Message.Chunk kids = 4;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_kids_size();
      for (const auto& msg : this_._internal_kids()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .AstroidMessages.Message.Chunk siblings = 5;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_siblings_size();
      for (const auto& msg : this_._internal_siblings()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // string mime_type = 6;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_mime_type().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mime_type());
      }
    }
    // string content = 10;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_content().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_content());
      }
    }
    // string sid = 13;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!this_._internal_sid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_sid());
      }
    }
    // string filename = 14;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!this_._internal_filename().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_filename());
      }
    }
    // string human_size = 16;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (!this_._internal_human_size().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_human_size());
      }
    }
    // string thumbnail = 17;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (!this_._internal_thumbnail().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_thumbnail());
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000ff00U)) {
    // string cid = 22;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (!this_._internal_cid().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_cid());
      }
    }
    // .AstroidMessages.Message.Chunk.Signature signature = 19;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      total_size += 2 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.signature_);
    }
    // .AstroidMessages.Message.Chunk.Encryption encryption = 20;
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      total_size += 2 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.encryption_);
    }
    // int32 id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      if (this_._internal_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_id());
      }
    }
    // bool viewable = 2;
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      if (this_._internal_viewable() != 0) {
        total_size += 2;
      }
    }
    // bool preferred = 3;
    if (CheckHasBit(cached_has_bits, 0x00002000U)) {
      if (this_._internal_preferred() != 0) {
        total_size += 2;
      }
    }
    // bool attachment = 7;
    if (CheckHasBit(cached_has_bits, 0x00004000U)) {
      if (this_._internal_attachment() != 0) {
        total_size += 2;
      }
    }
    // bool is_encrypted = 8;
    if (CheckHasBit(cached_has_bits, 0x00008000U)) {
      if (this_._internal_is_encrypted() != 0) {
        total_size += 2;
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x003f0000U)) {
    // bool is_signed = 9;
    if (CheckHasBit(cached_has_bits, 0x00010000U)) {
      if (this_._internal_is_signed() != 0) {
        total_size += 2;
      }
    }
    // bool sibling = 11;
    if (CheckHasBit(cached_has_bits, 0x00020000U)) {
      if (this_._internal_sibling() != 0) {
        total_size += 2;
      }
    }
    // bool use = 12;
    if (CheckHasBit(cached_has_bits, 0x00040000U)) {
      if (this_._internal_use() != 0) {
        total_size += 2;
      }
    }
    // bool focusable = 18;
    if (CheckHasBit(cached_has_bits, 0x00080000U)) {
      if (this_._internal_focusable() != 0) {
        total_size += 3;
      }
    }
    // int32 size = 15;
    if (CheckHasBit(cached_has_bits, 0x00100000U)) {
      if (this_._internal_size() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_size());
      }
    }
    // int32 crypto_id = 21;
    if (CheckHasBit(cached_has_bits, 0x00200000U)) {
      if (this_._internal_crypto_id() != 0) {
        total_size += 2 + ::_pbi::WireFormatLite::Int32Size(
                                        this_._internal_crypto_id());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Message_Chunk::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Message_Chunk*>(&to_msg);
  auto& from = static_cast<const Message_Chunk&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Message.Chunk)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_kids()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_kids());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_siblings()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_siblings());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_mime_type().empty()) {
        _this->_internal_set_mime_type(from._internal_mime_type());
      } else {
        if (_this->_impl_.mime_type_.IsDefault()) {
          _this->_internal_set_mime_type("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_content().empty()) {
        _this->_internal_set_content(from._internal_content());
      } else {
        if (_this->_impl_.content_.IsDefault()) {
          _this->_internal_set_content("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!from._internal_sid().empty()) {
        _this->_internal_set_sid(from._internal_sid());
      } else {
        if (_this->_impl_.sid_.IsDefault()) {
          _this->_internal_set_sid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!from._internal_filename().empty()) {
        _this->_internal_set_filename(from._internal_filename());
      } else {
        if (_this->_impl_.filename_.IsDefault()) {
          _this->_internal_set_filename("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (!from._internal_human_size().empty()) {
        _this->_internal_set_human_size(from._internal_human_size());
      } else {
        if (_this->_impl_.human_size_.IsDefault()) {
          _this->_internal_set_human_size("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (!from._internal_thumbnail().empty()) {
        _this->_internal_set_thumbnail(from._internal_thumbnail());
      } else {
        if (_this->_impl_.thumbnail_.IsDefault()) {
          _this->_internal_set_thumbnail("");
        }
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000ff00U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (!from._internal_cid().empty()) {
        _this->_internal_set_cid(from._internal_cid());
      } else {
        if (_this->_impl_.cid_.IsDefault()) {
          _this->_internal_set_cid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      ABSL_DCHECK(from._impl_.signature_ != nullptr);
      if (_this->_impl_.signature_ == nullptr) {
        _this->_impl_.signature_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.signature_);
      } else {
        _this->_impl_.signature_->MergeFrom(*from._impl_.signature_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      ABSL_DCHECK(from._impl_.encryption_ != nullptr);
      if (_this->_impl_.encryption_ == nullptr) {
        _this->_impl_.encryption_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.encryption_);
      } else {
        _this->_impl_.encryption_->MergeFrom(*from._impl_.encryption_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      if (from._internal_id() != 0) {
        _this->_impl_.id_ = from._impl_.id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      if (from._internal_viewable() != 0) {
        _this->_impl_.viewable_ = from._impl_.viewable_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00002000U)) {
      if (from._internal_preferred() != 0) {
        _this->_impl_.preferred_ = from._impl_.preferred_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00004000U)) {
      if (from._internal_attachment() != 0) {
        _this->_impl_.attachment_ = from._impl_.attachment_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00008000U)) {
      if (from._internal_is_encrypted() != 0) {
        _this->_impl_.is_encrypted_ = from._impl_.is_encrypted_;
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x003f0000U)) {
    if (CheckHasBit(cached_has_bits, 0x00010000U)) {
      if (from._internal_is_signed() != 0) {
        _this->_impl_.is_signed_ = from._impl_.is_signed_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00020000U)) {
      if (from._internal_sibling() != 0) {
        _this->_impl_.sibling_ = from._impl_.sibling_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00040000U)) {
      if (from._internal_use() != 0) {
        _this->_impl_.use_ = from._impl_.use_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00080000U)) {
      if (from._internal_focusable() != 0) {
        _this->_impl_.focusable_ = from._impl_.focusable_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00100000U)) {
      if (from._internal_size() != 0) {
        _this->_impl_.size_ = from._impl_.size_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00200000U)) {
      if (from._internal_crypto_id() != 0) {
        _this->_impl_.crypto_id_ = from._impl_.crypto_id_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Message_Chunk::CopyFrom(const Message_Chunk& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Message.Chunk)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Message_Chunk::InternalSwap(Message_Chunk* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.kids_.InternalSwap(&other->_impl_.kids_);
  _impl_.siblings_.InternalSwap(&other->_impl_.siblings_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mime_type_, &other->_impl_.mime_type_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.content_, &other->_impl_.content_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.sid_, &other->_impl_.sid_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.filename_, &other->_impl_.filename_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.human_size_, &other->_impl_.human_size_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.thumbnail_, &other->_impl_.thumbnail_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.cid_, &other->_impl_.cid_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.crypto_id_)
      + sizeof(Message_Chunk::_impl_.crypto_id_)
      - PROTOBUF_FIELD_OFFSET(Message_Chunk, _impl_.signature_)>(
          reinterpret_cast<char*>(&_impl_.signature_),
          reinterpret_cast<char*>(&other->_impl_.signature_));
}

::google::protobuf::Metadata Message_Chunk::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Message::Message(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.Message)
}
PROTOBUF_NDEBUG_INLINE Message::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::Message& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        tags_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.tags_)>()
          , from.tags_
        }
        ,
        mime_messages_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.mime_messages_)>()
          , from.mime_messages_
        }
        ,
        attachments_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::Message,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.attachments_)>()
          , from.attachments_
        }
        ,
        mid_(arena, from.mid_),
        date_pretty_(arena, from.date_pretty_),
        date_verbose_(arena, from.date_verbose_),
        subject_(arena, from.subject_),
        gravatar_(arena, from.gravatar_),
        in_reply_to_(arena, from.in_reply_to_),
        preview_(arena, from.preview_),
        tag_string_(arena, from.tag_string_) {}

Message::Message(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Message& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Message_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Message* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.sender_ = (CheckHasBit(cached_has_bits, 0x00000800U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.sender_)
                : nullptr;
  _impl_.to_ = (CheckHasBit(cached_has_bits, 0x00001000U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.to_)
                : nullptr;
  _impl_.cc_ = (CheckHasBit(cached_has_bits, 0x00002000U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.cc_)
                : nullptr;
  _impl_.bcc_ = (CheckHasBit(cached_has_bits, 0x00004000U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bcc_)
                : nullptr;
  _impl_.reply_to_ = (CheckHasBit(cached_has_bits, 0x00008000U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.reply_to_)
                : nullptr;
  _impl_.root_ = (CheckHasBit(cached_has_bits, 0x00010000U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.root_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, level_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, level_),
           offsetof(Impl_, different_subject_) -
               offsetof(Impl_, level_) +
               sizeof(Impl_::different_subject_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.Message)
}
PROTOBUF_NDEBUG_INLINE Message::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        tags_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.tags_)>()
         }
        ,
        mime_messages_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.mime_messages_)>()
         }
        ,
        attachments_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::Message,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::Message, _impl_.attachments_)>()
         }
        ,
        mid_(arena),
        date_pretty_(arena),
        date_verbose_(arena),
        subject_(arena),
        gravatar_(arena),
        in_reply_to_(arena),
        preview_(arena),
        tag_string_(arena) {}

inline void Message::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, sender_),
           0,
           offsetof(Impl_, different_subject_) -
               offsetof(Impl_, sender_) +
               sizeof(Impl_::different_subject_));
}
Message::~Message() {
  // @@protoc_insertion_point(destructor:AstroidMessages.Message)
  SharedDtor(*this);
}
inline void Message::SharedDtor(MessageLite& self) {
  Message& this_ = static_cast<Message&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.date_pretty_.Destroy();
  this_._impl_.date_verbose_.Destroy();
  this_._impl_.subject_.Destroy();
  this_._impl_.gravatar_.Destroy();
  this_._impl_.in_reply_to_.Destroy();
  this_._impl_.preview_.Destroy();
  this_._impl_.tag_string_.Destroy();
  delete this_._impl_.sender_;
  delete this_._impl_.to_;
  delete this_._impl_.cc_;
  delete this_._impl_.bcc_;
  delete this_._impl_.reply_to_;
  delete this_._impl_.root_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Message_class_data_ =
        Message::InternalGenerateClassData_(Message_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Message_class_data_.tc_table);
  return Message_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Message::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Message_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Message_globals_));
  return Message_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Message::ParseTableT_
    Message::_table_ =
        Message::InternalGenerateParseTable_(Message_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Message::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.Message)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.tags_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.mime_messages_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.attachments_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.mid_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.date_pretty_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _impl_.date_verbose_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _impl_.subject_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _impl_.gravatar_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000ff00U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      _impl_.in_reply_to_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      _impl_.preview_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      _impl_.tag_string_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      ABSL_DCHECK(_impl_.sender_ != nullptr);
      _impl_.sender_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      ABSL_DCHECK(_impl_.to_ != nullptr);
      _impl_.to_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00002000U)) {
      ABSL_DCHECK(_impl_.cc_ != nullptr);
      _impl_.cc_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00004000U)) {
      ABSL_DCHECK(_impl_.bcc_ != nullptr);
      _impl_.bcc_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00008000U)) {
      ABSL_DCHECK(_impl_.reply_to_ != nullptr);
      _impl_.reply_to_->Clear();
    }
  }
  if (CheckHasBit(cached_has_bits, 0x00010000U)) {
    ABSL_DCHECK(_impl_.root_ != nullptr);
    _impl_.root_->Clear();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x001e0000U)) {
    ::memset(&_impl_.level_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.different_subject_) -
        reinterpret_cast<char*>(&_impl_.level_)) + sizeof(_impl_.different_subject_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Message::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Message& this_ = static_cast<const Message&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Message::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Message& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.Message)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string mid = 1;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.mid");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // .AstroidMessages.Address sender = 2;
  if (CheckHasBit(cached_has_bits, 0x00000800U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.sender_, this_._impl_.sender_->GetCachedSize(), target,
        stream);
  }

  // .AstroidMessages.AddressList to = 3;
  if (CheckHasBit(cached_has_bits, 0x00001000U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.to_, this_._impl_.to_->GetCachedSize(), target,
        stream);
  }

  // .AstroidMessages.AddressList cc = 4;
  if (CheckHasBit(cached_has_bits, 0x00002000U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        4, *this_._impl_.cc_, this_._impl_.cc_->GetCachedSize(), target,
        stream);
  }

  // .AstroidMessages.AddressList bcc = 5;
  if (CheckHasBit(cached_has_bits, 0x00004000U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        5, *this_._impl_.bcc_, this_._impl_.bcc_->GetCachedSize(), target,
        stream);
  }

  // string date_pretty = 6;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (!this_._internal_date_pretty().empty()) {
      const ::std::string& _s = this_._internal_date_pretty();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.date_pretty");
      target = stream->WriteStringMaybeAliased(6, _s, target);
    }
  }

  // string date_verbose = 7;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (!this_._internal_date_verbose().empty()) {
      const ::std::string& _s = this_._internal_date_verbose();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.date_verbose");
      target = stream->WriteStringMaybeAliased(7, _s, target);
    }
  }

  // string subject = 9;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (!this_._internal_subject().empty()) {
      const ::std::string& _s = this_._internal_subject();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.subject");
      target = stream->WriteStringMaybeAliased(9, _s, target);
    }
  }

  // repeated string tags = 10;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_tags_size(); i < n; ++i) {
      const auto& s = this_._internal_tags().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.tags");
      target = stream->WriteString(10, s, target);
    }
  }

  // string gravatar = 11;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    if (!this_._internal_gravatar().empty()) {
      const ::std::string& _s = this_._internal_gravatar();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.gravatar");
      target = stream->WriteStringMaybeAliased(11, _s, target);
    }
  }

  // bool missing_content = 13;
  if (CheckHasBit(cached_has_bits, 0x00040000U)) {
    if (this_._internal_missing_content() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          13, this_._internal_missing_content(), target);
    }
  }

  // bool patch = 14;
  if (CheckHasBit(cached_has_bits, 0x00080000U)) {
    if (this_._internal_patch() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          14, this_._internal_patch(), target);
    }
  }

  // int32 level = 15;
  if (CheckHasBit(cached_has_bits, 0x00020000U)) {
    if (this_._internal_level() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<15>(
              stream, this_._internal_level(), target);
    }
  }

  // string in_reply_to = 16;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    if (!this_._internal_in_reply_to().empty()) {
      const ::std::string& _s = this_._internal_in_reply_to();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.in_reply_to");
      target = stream->WriteStringMaybeAliased(16, _s, target);
    }
  }

  // string preview = 17;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    if (!this_._internal_preview().empty()) {
      const ::std::string& _s = this_._internal_preview();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.preview");
      target = stream->WriteStringMaybeAliased(17, _s, target);
    }
  }

  // repeated .AstroidMessages.Message.Chunk mime_messages = 18;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_mime_messages_size());
         i < n; i++) {
      const auto& repfield = this_._internal_mime_messages().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              18, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .AstroidMessages.Message.Chunk attachments = 19;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_attachments_size());
         i < n; i++) {
      const auto& repfield = this_._internal_attachments().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              19, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // .AstroidMessages.Address reply_to = 20;
  if (CheckHasBit(cached_has_bits, 0x00008000U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        20, *this_._impl_.reply_to_, this_._impl_.reply_to_->GetCachedSize(), target,
        stream);
  }

  // string tag_string = 21;
  if (CheckHasBit(cached_has_bits, 0x00000400U)) {
    if (!this_._internal_tag_string().empty()) {
      const ::std::string& _s = this_._internal_tag_string();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.Message.tag_string");
      target = stream->WriteStringMaybeAliased(21, _s, target);
    }
  }

  // bool different_subject = 22;
  if (CheckHasBit(cached_has_bits, 0x00100000U)) {
    if (this_._internal_different_subject() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          22, this_._internal_different_subject(), target);
    }
  }

  // .AstroidMessages.Message.Chunk root = 23;
  if (CheckHasBit(cached_has_bits, 0x00010000U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        23, *this_._impl_.root_, this_._impl_.root_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.Message)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Message::ByteSizeLong(const MessageLite& base) {
  const Message& this_ = static_cast<const Message&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Message::ByteSizeLong() const {
  const Message& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.Message)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated string tags = 10;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_tags().size());
      for (int i = 0, n = this_._internal_tags().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_tags().Get(i));
      }
    }
    // repeated .AstroidMessages.Message.Chunk mime_messages = 18;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 2UL * this_._internal_mime_messages_size();
      for (const auto& msg : this_._internal_mime_messages()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .AstroidMessages.Message.Chunk attachments = 19;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 2UL * this_._internal_attachments_size();
      for (const auto& msg : this_._internal_attachments()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // string mid = 1;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // string date_pretty = 6;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!this_._internal_date_pretty().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_date_pretty());
      }
    }
    // string date_verbose = 7;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!this_._internal_date_verbose().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_date_verbose());
      }
    }
    // string subject = 9;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (!this_._internal_subject().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_subject());
      }
    }
    // string gravatar = 11;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (!this_._internal_gravatar().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_gravatar());
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000ff00U)) {
    // string in_reply_to = 16;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (!this_._internal_in_reply_to().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_in_reply_to());
      }
    }
    // string preview = 17;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (!this_._internal_preview().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_preview());
      }
    }
    // string tag_string = 21;
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      if (!this_._internal_tag_string().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_tag_string());
      }
    }
    // .AstroidMessages.Address sender = 2;
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.sender_);
    }
    // .AstroidMessages.AddressList to = 3;
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.to_);
    }
    // .AstroidMessages.AddressList cc = 4;
    if (CheckHasBit(cached_has_bits, 0x00002000U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.cc_);
    }
    // .AstroidMessages.AddressList bcc = 5;
    if (CheckHasBit(cached_has_bits, 0x00004000U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.bcc_);
    }
    // .AstroidMessages.Address reply_to = 20;
    if (CheckHasBit(cached_has_bits, 0x00008000U)) {
      total_size += 2 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.reply_to_);
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x001f0000U)) {
    // .AstroidMessages.Message.Chunk root = 23;
    if (CheckHasBit(cached_has_bits, 0x00010000U)) {
      total_size += 2 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.root_);
    }
    // int32 level = 15;
    if (CheckHasBit(cached_has_bits, 0x00020000U)) {
      if (this_._internal_level() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_level());
      }
    }
    // bool missing_content = 13;
    if (CheckHasBit(cached_has_bits, 0x00040000U)) {
      if (this_._internal_missing_content() != 0) {
        total_size += 2;
      }
    }
    // bool patch = 14;
    if (CheckHasBit(cached_has_bits, 0x00080000U)) {
      if (this_._internal_patch() != 0) {
        total_size += 2;
      }
    }
    // bool different_subject = 22;
    if (CheckHasBit(cached_has_bits, 0x00100000U)) {
      if (this_._internal_different_subject() != 0) {
        total_size += 3;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Message::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Message*>(&to_msg);
  auto& from = static_cast<const Message&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.Message)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_tags()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_tags());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_mime_messages()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_mime_messages());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_attachments()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_attachments());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!from._internal_date_pretty().empty()) {
        _this->_internal_set_date_pretty(from._internal_date_pretty());
      } else {
        if (_this->_impl_.date_pretty_.IsDefault()) {
          _this->_internal_set_date_pretty("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!from._internal_date_verbose().empty()) {
        _this->_internal_set_date_verbose(from._internal_date_verbose());
      } else {
        if (_this->_impl_.date_verbose_.IsDefault()) {
          _this->_internal_set_date_verbose("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (!from._internal_subject().empty()) {
        _this->_internal_set_subject(from._internal_subject());
      } else {
        if (_this->_impl_.subject_.IsDefault()) {
          _this->_internal_set_subject("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (!from._internal_gravatar().empty()) {
        _this->_internal_set_gravatar(from._internal_gravatar());
      } else {
        if (_this->_impl_.gravatar_.IsDefault()) {
          _this->_internal_set_gravatar("");
        }
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000ff00U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (!from._internal_in_reply_to().empty()) {
        _this->_internal_set_in_reply_to(from._internal_in_reply_to());
      } else {
        if (_this->_impl_.in_reply_to_.IsDefault()) {
          _this->_internal_set_in_reply_to("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (!from._internal_preview().empty()) {
        _this->_internal_set_preview(from._internal_preview());
      } else {
        if (_this->_impl_.preview_.IsDefault()) {
          _this->_internal_set_preview("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      if (!from._internal_tag_string().empty()) {
        _this->_internal_set_tag_string(from._internal_tag_string());
      } else {
        if (_this->_impl_.tag_string_.IsDefault()) {
          _this->_internal_set_tag_string("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      ABSL_DCHECK(from._impl_.sender_ != nullptr);
      if (_this->_impl_.sender_ == nullptr) {
        _this->_impl_.sender_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.sender_);
      } else {
        _this->_impl_.sender_->MergeFrom(*from._impl_.sender_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      ABSL_DCHECK(from._impl_.to_ != nullptr);
      if (_this->_impl_.to_ == nullptr) {
        _this->_impl_.to_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.to_);
      } else {
        _this->_impl_.to_->MergeFrom(*from._impl_.to_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00002000U)) {
      ABSL_DCHECK(from._impl_.cc_ != nullptr);
      if (_this->_impl_.cc_ == nullptr) {
        _this->_impl_.cc_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.cc_);
      } else {
        _this->_impl_.cc_->MergeFrom(*from._impl_.cc_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00004000U)) {
      ABSL_DCHECK(from._impl_.bcc_ != nullptr);
      if (_this->_impl_.bcc_ == nullptr) {
        _this->_impl_.bcc_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bcc_);
      } else {
        _this->_impl_.bcc_->MergeFrom(*from._impl_.bcc_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00008000U)) {
      ABSL_DCHECK(from._impl_.reply_to_ != nullptr);
      if (_this->_impl_.reply_to_ == nullptr) {
        _this->_impl_.reply_to_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.reply_to_);
      } else {
        _this->_impl_.reply_to_->MergeFrom(*from._impl_.reply_to_);
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x001f0000U)) {
    if (CheckHasBit(cached_has_bits, 0x00010000U)) {
      ABSL_DCHECK(from._impl_.root_ != nullptr);
      if (_this->_impl_.root_ == nullptr) {
        _this->_impl_.root_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.root_);
      } else {
        _this->_impl_.root_->MergeFrom(*from._impl_.root_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00020000U)) {
      if (from._internal_level() != 0) {
        _this->_impl_.level_ = from._impl_.level_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00040000U)) {
      if (from._internal_missing_content() != 0) {
        _this->_impl_.missing_content_ = from._impl_.missing_content_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00080000U)) {
      if (from._internal_patch() != 0) {
        _this->_impl_.patch_ = from._impl_.patch_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00100000U)) {
      if (from._internal_different_subject() != 0) {
        _this->_impl_.different_subject_ = from._impl_.different_subject_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Message::CopyFrom(const Message& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.Message)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Message::InternalSwap(Message* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.tags_.InternalSwap(&other->_impl_.tags_);
  _impl_.mime_messages_.InternalSwap(&other->_impl_.mime_messages_);
  _impl_.attachments_.InternalSwap(&other->_impl_.attachments_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.date_pretty_, &other->_impl_.date_pretty_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.date_verbose_, &other->_impl_.date_verbose_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.subject_, &other->_impl_.subject_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.gravatar_, &other->_impl_.gravatar_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.in_reply_to_, &other->_impl_.in_reply_to_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.preview_, &other->_impl_.preview_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.tag_string_, &other->_impl_.tag_string_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Message, _impl_.different_subject_)
      + sizeof(Message::_impl_.different_subject_)
      - PROTOBUF_FIELD_OFFSET(Message, _impl_.sender_)>(
          reinterpret_cast<char*>(&_impl_.sender_),
          reinterpret_cast<char*>(&other->_impl_.sender_));
}

::google::protobuf::Metadata Message::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

State_MessageState_Element::State_MessageState_Element(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_MessageState_Element_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.State.MessageState.Element)
}
PROTOBUF_NDEBUG_INLINE State_MessageState_Element::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::State_MessageState_Element& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        sid_(arena, from.sid_) {}

State_MessageState_Element::State_MessageState_Element(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const State_MessageState_Element& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_MessageState_Element_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  State_MessageState_Element* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, type_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, type_),
           offsetof(Impl_, focusable_) -
               offsetof(Impl_, type_) +
               sizeof(Impl_::focusable_));

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.State.MessageState.Element)
}
PROTOBUF_NDEBUG_INLINE State_MessageState_Element::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        sid_(arena) {}

inline void State_MessageState_Element::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, type_),
           0,
           offsetof(Impl_, focusable_) -
               offsetof(Impl_, type_) +
               sizeof(Impl_::focusable_));
}
State_MessageState_Element::~State_MessageState_Element() {
  // @@protoc_insertion_point(destructor:AstroidMessages.State.MessageState.Element)
  SharedDtor(*this);
}
inline void State_MessageState_Element::SharedDtor(MessageLite& self) {
  State_MessageState_Element& this_ = static_cast<State_MessageState_Element&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.sid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull State_MessageState_Element_class_data_ =
        State_MessageState_Element::InternalGenerateClassData_(State_MessageState_Element_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State_MessageState_Element::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_MessageState_Element_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(State_MessageState_Element_class_data_.tc_table);
  return State_MessageState_Element_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State_MessageState_Element::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_MessageState_Element_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&State_MessageState_Element_globals_));
  return State_MessageState_Element_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const State_MessageState_Element::ParseTableT_
    State_MessageState_Element::_table_ =
        State_MessageState_Element::InternalGenerateParseTable_(State_MessageState_Element_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void State_MessageState_Element::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.State.MessageState.Element)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.sid_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000000eU)) {
    ::memset(&_impl_.type_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.focusable_) -
        reinterpret_cast<char*>(&_impl_.type_)) + sizeof(_impl_.focusable_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL State_MessageState_Element::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const State_MessageState_Element& this_ = static_cast<const State_MessageState_Element&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL State_MessageState_Element::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const State_MessageState_Element& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.State.MessageState.Element)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .AstroidMessages.State.MessageState.Element.Type type = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_type() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_type(), target);
    }
  }

  // int32 id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<2>(
              stream, this_._internal_id(), target);
    }
  }

  // string sid = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_sid().empty()) {
      const ::std::string& _s = this_._internal_sid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.State.MessageState.Element.sid");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // bool focusable = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_focusable() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          4, this_._internal_focusable(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.State.MessageState.Element)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t State_MessageState_Element::ByteSizeLong(const MessageLite& base) {
  const State_MessageState_Element& this_ = static_cast<const State_MessageState_Element&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t State_MessageState_Element::ByteSizeLong() const {
  const State_MessageState_Element& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.State.MessageState.Element)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // string sid = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_sid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_sid());
      }
    }
    // .AstroidMessages.State.MessageState.Element.Type type = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_type() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_type());
      }
    }
    // int32 id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_id());
      }
    }
    // bool focusable = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_focusable() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void State_MessageState_Element::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<State_MessageState_Element*>(&to_msg);
  auto& from = static_cast<const State_MessageState_Element&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.State.MessageState.Element)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_sid().empty()) {
        _this->_internal_set_sid(from._internal_sid());
      } else {
        if (_this->_impl_.sid_.IsDefault()) {
          _this->_internal_set_sid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_type() != 0) {
        _this->_impl_.type_ = from._impl_.type_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_id() != 0) {
        _this->_impl_.id_ = from._impl_.id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_focusable() != 0) {
        _this->_impl_.focusable_ = from._impl_.focusable_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void State_MessageState_Element::CopyFrom(const State_MessageState_Element& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.State.MessageState.Element)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void State_MessageState_Element::InternalSwap(State_MessageState_Element* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.sid_, &other->_impl_.sid_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.focusable_)
      + sizeof(State_MessageState_Element::_impl_.focusable_)
      - PROTOBUF_FIELD_OFFSET(State_MessageState_Element, _impl_.type_)>(
          reinterpret_cast<char*>(&_impl_.type_),
          reinterpret_cast<char*>(&other->_impl_.type_));
}

::google::protobuf::Metadata State_MessageState_Element::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

State_MessageState::State_MessageState(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_MessageState_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.State.MessageState)
}
PROTOBUF_NDEBUG_INLINE State_MessageState::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::State_MessageState& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mid_(arena, from.mid_),
        elements_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::State_MessageState,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.elements_)>()
          , from.elements_
        }
     {}

State_MessageState::State_MessageState(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const State_MessageState& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_MessageState_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  State_MessageState* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.level_ = from._impl_.level_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.State.MessageState)
}
PROTOBUF_NDEBUG_INLINE State_MessageState::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mid_(arena),
        elements_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::State_MessageState,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::State_MessageState, _impl_.elements_)>()
         }
     {}

inline void State_MessageState::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.level_ = {};
}
State_MessageState::~State_MessageState() {
  // @@protoc_insertion_point(destructor:AstroidMessages.State.MessageState)
  SharedDtor(*this);
}
inline void State_MessageState::SharedDtor(MessageLite& self) {
  State_MessageState& this_ = static_cast<State_MessageState&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.mid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull State_MessageState_class_data_ =
        State_MessageState::InternalGenerateClassData_(State_MessageState_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State_MessageState::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_MessageState_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(State_MessageState_class_data_.tc_table);
  return State_MessageState_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State_MessageState::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_MessageState_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&State_MessageState_globals_));
  return State_MessageState_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const State_MessageState::ParseTableT_
    State_MessageState::_table_ =
        State_MessageState::InternalGenerateParseTable_(State_MessageState_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void State_MessageState::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.State.MessageState)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.mid_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    _impl_.level_ = 0;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.elements_.Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL State_MessageState::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const State_MessageState& this_ = static_cast<const State_MessageState&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL State_MessageState::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const State_MessageState& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.State.MessageState)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string mid = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_mid().empty()) {
      const ::std::string& _s = this_._internal_mid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "AstroidMessages.State.MessageState.mid");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // repeated .AstroidMessages.State.MessageState.Element elements = 5;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_elements_size());
         i < n; i++) {
      const auto& repfield = this_._internal_elements().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              5, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // int32 level = 6;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_level() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<6>(
              stream, this_._internal_level(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.State.MessageState)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t State_MessageState::ByteSizeLong(const MessageLite& base) {
  const State_MessageState& this_ = static_cast<const State_MessageState&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t State_MessageState::ByteSizeLong() const {
  const State_MessageState& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.State.MessageState)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string mid = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_mid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mid());
      }
    }
    // int32 level = 6;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_level() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_level());
      }
    }
    // repeated .AstroidMessages.State.MessageState.Element elements = 5;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_elements_size();
      for (const auto& msg : this_._internal_elements()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void State_MessageState::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<State_MessageState*>(&to_msg);
  auto& from = static_cast<const State_MessageState&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.State.MessageState)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_mid().empty()) {
        _this->_internal_set_mid(from._internal_mid());
      } else {
        if (_this->_impl_.mid_.IsDefault()) {
          _this->_internal_set_mid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_level() != 0) {
        _this->_impl_.level_ = from._impl_.level_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_elements()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_elements());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void State_MessageState::CopyFrom(const State_MessageState& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.State.MessageState)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void State_MessageState::InternalSwap(State_MessageState* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mid_, &other->_impl_.mid_, arena);
  swap(_impl_.level_, other->_impl_.level_);
  _impl_.elements_.InternalSwap(&other->_impl_.elements_);
}

::google::protobuf::Metadata State_MessageState::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

State::State(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.State)
}
PROTOBUF_NDEBUG_INLINE State::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::State& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        messages_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::AstroidMessages::State,
              PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_.messages_)>()
          , from.messages_
        }
     {}

State::State(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const State& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, State_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  State* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.edit_mode_ = from._impl_.edit_mode_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.State)
}
PROTOBUF_NDEBUG_INLINE State::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        messages_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::AstroidMessages::State,
            PROTOBUF_FIELD_OFFSET(::AstroidMessages::State, _impl_.messages_)>()
         }
     {}

inline void State::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.edit_mode_ = {};
}
State::~State() {
  // @@protoc_insertion_point(destructor:AstroidMessages.State)
  SharedDtor(*this);
}
inline void State::SharedDtor(MessageLite& self) {
  State& this_ = static_cast<State&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull State_class_data_ =
        State::InternalGenerateClassData_(State_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(State_class_data_.tc_table);
  return State_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
State::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&State_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&State_globals_));
  return State_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const State::ParseTableT_
    State::_table_ =
        State::InternalGenerateParseTable_(State_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void State::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.State)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.messages_.Clear();
  }
  _impl_.edit_mode_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL State::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const State& this_ = static_cast<const State&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL State::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const State& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.State)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .AstroidMessages.State.MessageState messages = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_messages_size());
         i < n; i++) {
      const auto& repfield = this_._internal_messages().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // bool edit_mode = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_edit_mode() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          3, this_._internal_edit_mode(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.State)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t State::ByteSizeLong(const MessageLite& base) {
  const State& this_ = static_cast<const State&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t State::ByteSizeLong() const {
  const State& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.State)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated .AstroidMessages.State.MessageState messages = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_messages_size();
      for (const auto& msg : this_._internal_messages()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // bool edit_mode = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_edit_mode() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void State::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<State*>(&to_msg);
  auto& from = static_cast<const State&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.State)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_messages()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_messages());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_edit_mode() != 0) {
        _this->_impl_.edit_mode_ = from._impl_.edit_mode_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void State::CopyFrom(const State& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.State)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void State::InternalSwap(State* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.messages_.InternalSwap(&other->_impl_.messages_);
  swap(_impl_.edit_mode_, other->_impl_.edit_mode_);
}

::google::protobuf::Metadata State::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

UpdateMessage::UpdateMessage(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateMessage_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.UpdateMessage)
}
PROTOBUF_NDEBUG_INLINE UpdateMessage::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::AstroidMessages::UpdateMessage& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

UpdateMessage::UpdateMessage(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const UpdateMessage& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateMessage_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  UpdateMessage* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.m_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.m_)
                : nullptr;
  _impl_.type_ = from._impl_.type_;

  // @@protoc_insertion_point(copy_constructor:AstroidMessages.UpdateMessage)
}
PROTOBUF_NDEBUG_INLINE UpdateMessage::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void UpdateMessage::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, m_),
           0,
           offsetof(Impl_, type_) -
               offsetof(Impl_, m_) +
               sizeof(Impl_::type_));
}
UpdateMessage::~UpdateMessage() {
  // @@protoc_insertion_point(destructor:AstroidMessages.UpdateMessage)
  SharedDtor(*this);
}
inline void UpdateMessage::SharedDtor(MessageLite& self) {
  UpdateMessage& this_ = static_cast<UpdateMessage&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.m_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull UpdateMessage_class_data_ =
        UpdateMessage::InternalGenerateClassData_(UpdateMessage_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateMessage::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateMessage_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(UpdateMessage_class_data_.tc_table);
  return UpdateMessage_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateMessage::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateMessage_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&UpdateMessage_globals_));
  return UpdateMessage_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const UpdateMessage::ParseTableT_
    UpdateMessage::_table_ =
        UpdateMessage::InternalGenerateParseTable_(UpdateMessage_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void UpdateMessage::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.UpdateMessage)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(_impl_.m_ != nullptr);
    _impl_.m_->Clear();
  }
  _impl_.type_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL UpdateMessage::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const UpdateMessage& this_ = static_cast<const UpdateMessage&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL UpdateMessage::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const UpdateMessage& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.UpdateMessage)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .AstroidMessages.Message m = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.m_, this_._impl_.m_->GetCachedSize(), target,
        stream);
  }

  // .AstroidMessages.UpdateMessage.Type type = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_type() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          2, this_._internal_type(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.UpdateMessage)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t UpdateMessage::ByteSizeLong(const MessageLite& base) {
  const UpdateMessage& this_ = static_cast<const UpdateMessage&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t UpdateMessage::ByteSizeLong() const {
  const UpdateMessage& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.UpdateMessage)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // .AstroidMessages.Message m = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.m_);
    }
    // .AstroidMessages.UpdateMessage.Type type = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_type() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_type());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void UpdateMessage::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<UpdateMessage*>(&to_msg);
  auto& from = static_cast<const UpdateMessage&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.UpdateMessage)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.m_ != nullptr);
      if (_this->_impl_.m_ == nullptr) {
        _this->_impl_.m_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.m_);
      } else {
        _this->_impl_.m_->MergeFrom(*from._impl_.m_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_type() != 0) {
        _this->_impl_.type_ = from._impl_.type_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void UpdateMessage::CopyFrom(const UpdateMessage& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.UpdateMessage)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void UpdateMessage::InternalSwap(UpdateMessage* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.type_)
      + sizeof(UpdateMessage::_impl_.type_)
      - PROTOBUF_FIELD_OFFSET(UpdateMessage, _impl_.m_)>(
          reinterpret_cast<char*>(&_impl_.m_),
          reinterpret_cast<char*>(&other->_impl_.m_));
}

::google::protobuf::Metadata UpdateMessage::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ClearMessage::ClearMessage(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ClearMessage_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:AstroidMessages.ClearMessage)
}
ClearMessage::ClearMessage(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ClearMessage& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ClearMessage_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE ClearMessage::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void ClearMessage::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.yes_ = {};
}
ClearMessage::~ClearMessage() {
  // @@protoc_insertion_point(destructor:AstroidMessages.ClearMessage)
  SharedDtor(*this);
}
inline void ClearMessage::SharedDtor(MessageLite& self) {
  ClearMessage& this_ = static_cast<ClearMessage&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ClearMessage_class_data_ =
        ClearMessage::InternalGenerateClassData_(ClearMessage_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ClearMessage::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ClearMessage_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ClearMessage_class_data_.tc_table);
  return ClearMessage_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ClearMessage::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ClearMessage_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ClearMessage_globals_));
  return ClearMessage_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ClearMessage::ParseTableT_
    ClearMessage::_table_ =
        ClearMessage::InternalGenerateParseTable_(ClearMessage_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ClearMessage::Clear() {
// @@protoc_insertion_point(message_clear_start:AstroidMessages.ClearMessage)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.yes_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ClearMessage::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ClearMessage& this_ = static_cast<const ClearMessage&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ClearMessage::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ClearMessage& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:AstroidMessages.ClearMessage)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bool yes = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_yes() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          1, this_._internal_yes(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:AstroidMessages.ClearMessage)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ClearMessage::ByteSizeLong(const MessageLite& base) {
  const ClearMessage& this_ = static_cast<const ClearMessage&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ClearMessage::ByteSizeLong() const {
  const ClearMessage& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:AstroidMessages.ClearMessage)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bool yes = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_yes() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ClearMessage::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ClearMessage*>(&to_msg);
  auto& from = static_cast<const ClearMessage&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:AstroidMessages.ClearMessage)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_yes() != 0) {
      _this->_impl_.yes_ = from._impl_.yes_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ClearMessage::CopyFrom(const ClearMessage& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:AstroidMessages.ClearMessage)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ClearMessage::InternalSwap(ClearMessage* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.yes_, other->_impl_.yes_);
}

::google::protobuf::Metadata ClearMessage::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace AstroidMessages
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_messages_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
